<?php

/*
+---------------------------------------------------------------------------+
| Max Media Manager v0.3                                                    |
| =================                                                         |
|                                                                           |
| Copyright (c) 2003-2005 m3 Media Services Limited                         |
| For contact details, see: http://www.m3.net/                              |
|                                                                           |
| This program is free software; you can redistribute it and/or modify      |
| it under the terms of the GNU General Public License as published by      |
| the Free Software Foundation; either version 2 of the License, or         |
| (at your option) any later version.                                       |
|                                                                           |
| This program is distributed in the hope that it will be useful,           |
| but WITHOUT ANY WARRANTY; without even the implied warranty of            |
| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             |
| GNU General Public License for more details.                              |
|                                                                           |
| You should have received a copy of the GNU General Public License         |
| along with this program; if not, write to the Free Software               |
| Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA |
+---------------------------------------------------------------------------+
$Id: Maintenance.php 3707 2005-09-30 16:40:21Z andrew $
*/

require_once MAX_PATH . '/lib/max/Dal/Common.php';

/**
 * The non-DB specific Data Access Layer (DAL) class for the Maintenance Engine.
 *
 * Includes DAL methods for both the Maintenance Statistics Engine and Maintenance
 * Priority Engine processes.
 *
 * @package    MaxDal
 * @author     James Floyd <james@m3.net>
 * @author     Andrew Hill <andrew@m3.net>
 * @author     Radek Maciaszek <radek@m3.net>
 */
class MAX_Dal_Maintenance extends MAX_Dal_Common 
{
    /**
     * The class constructor method.
     */
    function MAX_Dal_Maintenance()
    {
        $this->MAX_Dal_Common();
    }
    
    /**
     * A singleton method to obtain the same instance of the class.
     *
     * @access public
     */
    function &singleton()
    {
        static $instance;
        if (!is_a($instance, 'MAX_Dal_Maintenance')) {
            $instance = &new MAX_Dal_Maintenance();
        }
        return $instance;
    }
    
    /**
     * A method to store the last time that the maintenance priority script
     * ran.
     *
     * @param Date $start The time that the maintenance priority run started.
     * @param Date $end The time that the maintenance priority run ended.
     * @param Date $updateTo The end of the last operation interval ID that
     *                       has been updated.
     */
    function setMaintenancePriorityLastRunInfo($start, $end, $updateTo)
    {
        return parent::setScriptLastRunInfo($start, $end, $updateTo, 'log_maintenance_priority');
    }
    
    /**
     * A method to store the last time that the maintenance satistics script
     * ran.
     *
     * @param Date $start The time that the maintenance statistics run started.
     * @param Date $end The time that the maintenance statistics run ended.
     * @param Integer $type type of statistics run performed.
     * @param Date $updateTo The end of the last operation interval ID that
     *                       has been updated.
     * @param String $runTypeField name of DB field to hold $type value
     *                      currently 'adserver_run_type' or 'tracker_run_type'.
     */
    function setMaintenanceStatisticsLastRunInfo($start, $end, $type, $updateTo, $runTypeField)
    {
        if (empty($runTypeField)) {
            return PEAR::raiseError('$runTypeField parameter requires a value.', MAX_ERROR_INVALIDARGS);
        }
        
        $conf = $GLOBALS['_MAX']['CONF'];
        $oStartDateCopy = &new Date($start); // Clone for Date_Span::setFromDateDiff bug in PHP5
        $oEndDateCopy   = &new Date($end);   // Clone for Date_Span::setFromDateDiff bug in PHP5
        $duration = new Date_Span();
        $duration->setFromDateDiff($oStartDateCopy, $oEndDateCopy);
        
        $query = "
            INSERT INTO
                {$conf['table']['prefix']}{$conf['table']['log_maintenance_statistics']}
                (
                    start_run,
                    end_run,
                    duration,
                    {$runTypeField},
                    updated_to
                )
            VALUES
                (
                    '".$start->format('%Y-%m-%d %H:%M:%S')."',
                    '".$end->format('%Y-%m-%d %H:%M:%S')."',
                    ".$duration->toSeconds().",
                    $type,
                    '".$updateTo->format('%Y-%m-%d %H:%M:%S')."'
                )";
        MAX::debug('Logging the maintenance statistics run information', PEAR_LOG_DEBUG);
        return $this->dbh->query($query);
    }
    
    /**
     * A method to store the last maintenance satistics run report.
     * ran.
     *
     * @param String $report The report to be logged.
     */
    function setMaintenanceStatisticesRunReport($report)
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $query = "
            INSERT INTO
                {$conf['table']['prefix']}{$conf['table']['userlog']}
                (
                    timestamp,
                    usertype,
                    userid,
                    action,
                    object,
                    details
                )
            VALUES
                (
                    '".time()."',
                    '".phpAds_userMaintenance."',
                    0,
                    '".phpAds_actionBatchStatistics."',
                    0,
                    '".addslashes($report)."'
                )";
        MAX::debug('Logging the maintenance statistics run report', PEAR_LOG_DEBUG);
        return $this->dbh->query($query);
    }
    
    /**
     * A method to return the time that the maintenance priority scripts
     * were last run, and the operation interval that was in use at the time.
     *
     * @return array An array containing the time that represents the last time
     *               maintenance priority was last run (started). Returns an
     *               array of 2 NULLs if maintenance priority has never been run.
     */
    function getMaintenancePriorityLastRunInfo()
    {
        return parent::getScriptLastRunInfo('log_maintenance_priority', array('operation_interval'));
    }
    
    /**
     * A method to return the time that the maintenance statistics scripts
     * were last run, and the type of update that was done at the time.
     *
     * @return array An array containing the time that represents the time
     *               that the maintenance statistics were last updated to,
     *               index as "updated_to", and the type of update done,
     *               indexed as "adserver_run_type".
     */
    function getMaintenanceStatsLastRunInfo()
    {
        return parent::getScriptLastRunInfo('log_maintenance_statistics', array('adserver_run_type'));
    }
    
    /**
     * A method to get all placements (used to be campaigns).
     *
     * @param $fields   array An array of extra fields to select, if required.
     * @param $where    array An array of extra where statements to filter on, if required.
     *                        This would almost always be set, to limit the placements returned
     *                        to some sub-set of all the placements.
     * @param $joins    array An array of extra join statements to join with, if required.
     * @param $orderBy  array An array of extra order by statements to order with, if required.
     * @return array 
     */
    function getPlacements($fields = array(), $wheres = array(), $joins = array(), $orderBys = array())
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $query = array();
        $table             = $conf['table']['prefix'] . $conf['table']['campaigns'];
        $query['table']    = $table;
        $query['fields']   = array(
                                "$table.campaignid",
                                "$table.views",
                                "$table.clicks",
                                "$table.conversions",
                                "$table.expire",
                                "$table.target_impression",
                                "$table.target_click",
                                "$table.target_conversion",
                                "$table.priority"
                             );
        // Add the custom fields
        if (is_array($fields) && (count($fields) > 0)) {
            foreach ($fields as $field) {
                $query['fields'][] = "$field";
            }
        }
        // Add the custom where statements
        if (is_array($wheres) && (count($wheres) > 0)) {
            foreach ($wheres as $where) {
                $query['wheres'][] = $where;
            }
        }
        // Add the custom join statements
        if (is_array($joins) && (count($joins) > 0)) {
            foreach ($joins as $join) {
                $query['joins'][] = $join;
            }
        }
        // Add the custom order by statements
        if (is_array($orderBys) && (count($orderBys) > 0)) {
            foreach ($orderBys as $orderBy) {
                $query['orderBys'][] = $orderBy;
            }
        }
        return $this->_get($query);
    }
    
    /**
     * A method to get basic data about placements (previously campaigns). 
     *
     * @param integer $id The placement ID.
     * @return array An array containing the advertiser_id, placement_id,
     *               and name of the placement, if the placement is "active"
     *               or not (true/false), and the "num_children" value (the
     *               number of advertisements in the placement).
     */
    function getPlacementData($id)
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $query = array();
        $table             = $conf['table']['prefix'] . $conf['table']['campaigns'];
        $joinTable         = $conf['table']['prefix'] . $conf['table']['banners'];
        $query['table']    = $table;
        $query['fields']   = array(
                                "$table.clientid AS advertiser_id",
                                "$table.campaignid AS placement_id",
                                "$table.campaignname AS name",
                                "$table.active AS active",
                                "COUNT($joinTable.bannerid) AS num_children"
                             );
        $query['wheres']   = array(
                                array("$table.campaignid = $id", 'AND')
                             );
        $query['joins']    = array(
                                array($joinTable, "$table.campaignid = $joinTable.campaignid")
                             );
        $query['group']    = "advertiser_id, placement_id, name, active";
        return $this->_get($query);
    }
    
    /**
     * A method to get the details of the total number of advertisements
     * delivered, to date, for a specified placement.
     *
     * @param integer $id The placement ID.
     * @return array An array of arrays, with each containing the "placement_id",
     *               "sum_requests", "sum_views", "sum_clicks" and "sum_conversions"
     *               for that placement.
     */
    function getPlacementDeliveryToDate($id)
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $query = array();
        $table             = $conf['table']['prefix'] . $conf['table']['campaigns'];
        $joinTable1        = $conf['table']['prefix'] . $conf['table']['banners'];
        $joinTable2        = $conf['table']['prefix'] . $conf['table']['data_intermediate_ad'];
        $query['table']    = $table;
        $query['fields']   = array(
                                "SUM($joinTable2.requests) AS sum_requests",
                                "SUM($joinTable2.impressions) AS sum_views",
                                "SUM($joinTable2.clicks) AS sum_clicks",
                                "SUM($joinTable2.conversions) AS sum_conversions",
                                "$table.campaignid AS placement_id"
                             );
        $query['wheres']   = array(
                                array("$table.campaignid = $id", 'AND')
                             );
        $query['joins']    = array(
                                array($joinTable1, "$table.campaignid = $joinTable1.campaignid"),
                                array($joinTable2, "$joinTable1.bannerid = $joinTable2.ad_id")
                             );
        $query['group']    = "placement_id";
        return $this->_get($query);
    }
    
    /**
     * A method to get the details of the total number of advertisements
     * delivered, today, for a specified placement.
     *
     * @param integer $id The placement ID.
     * @param string $today A "CCYY-MM-DD" formatted representation of today's date.
     * @return array An array of arrays, with each containing the "placement_id",
     *               "sum_requests", "sum_views", "sum_clicks" and "sum_conversions"
     *               for that placement.
     */
    function getPlacementDeliveryToday($id, $today)
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $query = array();
        $table             = $conf['table']['prefix'] . $conf['table']['campaigns'];
        $joinTable1        = $conf['table']['prefix'] . $conf['table']['banners'];
        $joinTable2        = $conf['table']['prefix'] . $conf['table']['data_intermediate_ad'];
        $query['table']    = $table;
        $query['fields']   = array(
                                "SUM($joinTable2.requests) AS sum_requests",
                                "SUM($joinTable2.impressions) AS sum_views",
                                "SUM($joinTable2.clicks) AS sum_clicks",
                                "SUM($joinTable2.conversions) AS sum_conversions",
                                "$table.campaignid AS placement_id"
                             );
        $query['wheres']   = array(
                                array("$table.campaignid = $id", 'AND')
                             );
        $query['joins']    = array(
                                array($joinTable1, "$table.campaignid = $joinTable1.campaignid"),
                                array($joinTable2, "$joinTable1.bannerid = $joinTable2.ad_id AND $joinTable2.day = '$today'")
                             );
        $query['group']    = "placement_id";
        return $this->_get($query);
    }

    /**
     * A convenience method that combines the results of the getPlacementData and
     * (getPlacementDeliveryToDate or getPlacementDeliverToday) methods into one
     * result set.
     *
     * @param integer $id The placement ID.
     * @param boolean $todayOnly If true, only look at what has been delivered today
     *                           for the delivery statistics (ie. use getPlacementDeliverToday),
     *                           otherwise use the delivery statistics for the entire placement
     *                           lifetime (ie. use getPlacementDeliveryToDate).
     * @param string $today A "CCYY-MM-DD" formatted representation of today's date, only
     *                      required when $todayOnly is true.
     * @return array An array of the placement's details, containing the following:
     *
     * Array
     *     (
     *         [advertiser_id]   => 1
     *         [placement_id]    => 1
     *         [name]            => Placement name
     *         [active]          => t
     *         [num_children]    => 2
     *         [sum_requests]    => 0
     *         [sum_views]       => 0
     *         [sum_clicks]      => 0
     *         [sum_conversions] => 0
     *     )
     */
    function getPlacementStats($id, $todayOnly = false, $today = '')
    {
        $aTemp = $this->getPlacementData($id);
        if (is_array($aTemp[0]) && (count($aTemp[0]) > 0)) {
            $aPlacements = $aTemp[0];
        }
        if ($todayOnly) {
            $aPlacementsStats = $this->getPlacementDeliveryToday($id, $today);
        } else {
            $aPlacementsStats = $this->getPlacementDeliveryToDate($id);
        }
        $items = array(
            'sum_requests',
            'sum_views',
            'sum_clicks',
            'sum_conversions'
        );
        foreach ($items as $item) {
            if (!empty($aPlacementsStats[0][$item])) {
                $aPlacements[$item] = $aPlacementsStats[0][$item] ;
            } else {
                $aPlacements[$item] = 0;
            }
        }
        return $aPlacements;
    }
    
    /**
     * A method to get all zones that have had advertisements allocated to their
     * impression inventory.
     *
     * @access public
     * @return mixed An array of arrays, each one representing a zone/ad pair
     *               found, in zone ID order.
     */
    function &getAllZonesWithAllocInv()
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $query = array();
        $table             = 'tmp_ad_zone_impression';
        $query['table']    = $table;
        $query['fields']   = array(
                                "$table.zone_id",
                                "$table.ad_id",
                                "$table.impressions"
                             );
        $query['orderBys'] = array(
                                array("$table.zone_id", 'ASC'),
                                array("$table.ad_id", 'ASC')
                             );
        return $this->_get($query);
    }
    
    /**
     * A method to get the available impression inventory for all zones 
     * for the current operation interval, and the actual impressions
     * for the previous operation interval.
     *
     * Requires that a current day/time (as a Date object) be registered
     * with the ServiceLocator (as "now").
     *
     * Ignores the special zone with zone ID of 0 (zero), as this is a
     * reserved zone for storing statistics about directly selected ads,
     * and as a result, shouldn't be used during prioritisation.
     *
     * @access public
     * @return mixed An array of arrays, each one representing a zone found,
     *               in zone ID order, or false if no Date object registered
     *               with the ServiceLocator.
     */
    function &getAllZonesImpInv()
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $dbh = &MAX_DB::singleton();
        $oServiceLocator = &ServiceLocator::instance();
        $oDate = &$oServiceLocator->get('now');
        if (!$oDate) {
            return false;
        }
        $oOperationInterval = &OperationInterval::singleton();
        // Get the forecast impressions for the current operation interval
        $currentOpIntID = $oOperationInterval->getIntervalID($oDate);
        $aCurrentDates = $oOperationInterval->convertDateToOperationIntervalStartAndEndDates($oDate);
        $previousOptIntID = $oOperationInterval->getPreviousIntervalID($currentOpIntID);
        $aPreviousDates = $oOperationInterval->convertDateToPreviousOperationIntervalStartAndEndDates($oDate);
        $table = $conf['table']['prefix'] . $conf['table']['data_summary_zone_impression_history'];
        $query = "
            SELECT
                t1.zone_id AS zone_id,
                t1.forecast_impressions AS forecast_impressions,
                t2.actual_impressions AS actual_impressions
            FROM
                $table AS t1
            LEFT JOIN
                $table AS t2
            ON
                t1.zone_id = t2.zone_id
                AND t2.operation_interval = {$conf['maintenance']['operationInterval']}
                AND t2.operation_interval_id = $previousOptIntID
                AND t2.interval_start = '" . $aPreviousDates['start']->format('%Y-%m-%d %H:%M:%S') . "'
                AND t2.interval_end = '" . $aPreviousDates['end']->format('%Y-%m-%d %H:%M:%S') . "'
                AND t2.zone_id != 0
            WHERE
                t1.operation_interval = {$conf['maintenance']['operationInterval']}
                AND t1.operation_interval_id = $currentOpIntID
                AND t1.interval_start = '" . $aCurrentDates['start']->format('%Y-%m-%d %H:%M:%S') . "'
                AND t1.interval_end = '" . $aCurrentDates['end']->format('%Y-%m-%d %H:%M:%S') . "'
                AND t1.zone_id != 0
            ORDER BY
                t1.zone_id";
        return $dbh->getAll($query);
    }
    
    /**
     * A method to get all active advertisements that have one of:
     *  - A delivery limitation.
     *  - A user time-based delivery cap.
     *  - A user total-based delivery cap.
     *  - A user session-based delivery cap.
     *
     * @access public
     * @return mixed An array of arrays, each one representing an ad found,
     *               in advertisement ID order.
     */
    function &getAllDeliveryLimitedAds()
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $query = array();
        $table             = $conf['table']['prefix'] . $conf['table']['banners'];
        $joinTable         = $conf['table']['prefix'] . $conf['table']['campaigns'];
        $query['table']    = $table;
        $query['fields']   = array(
                                "$table.bannerid",
                                "$table.block",
                                "$table.capping",
                                "$table.session_capping",
                                "$table.compiledlimitation"
                             );
        $query['wheres']   = array(
                                array("$table.active = 't'", 'AND'),
                                array("($table.block > 0 OR $table.capping > 0 OR $table.session_capping > 0 OR " .
                                      "$table.compiledlimitation != 'true')", 'AND')
                             );
        $query['joins']    = array(
                                array($joinTable, "$table.campaignid = $joinTable.campaignid AND $joinTable.active = 't'")
                             );
        $query['orderBys'] = array(
                                array("$table.bannerid", 'ASC')
                             );
        return $this->_get($query);
    }
    
    /**
     * A method to get the previous operation interval's details of advertisement
     * delivery for all advertisements.
     *
     * Requires that a current day/time (as a Date object) be registered
     * with the ServiceLocator (as "now").
     *
     * @access public
     * @return mixed An array of arrays, each one representing an ad/zone 
     *               delivery information set, in ad/zone ID order, or false
     *               if no Date object registered with the ServiceLocator.
     */
    function &getLastOperationIntervalAdDeliveryInfo()
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $oServiceLocator = &ServiceLocator::instance();
        $oDate = &$oServiceLocator->get('now');
        if (!$oDate) {
            return false;
        }
        $oOperationInterval = &OperationInterval::singleton();
        $currentOperationIntervalID = $oOperationInterval->getIntervalID($oDate);
        $previousOptIntID = $oOperationInterval->getPreviousIntervalID($currentOperationIntervalID);
        $aDates = $oOperationInterval->convertDateToPreviousOperationIntervalStartAndEndDates($oDate);
        $query = array();
        $table             = $conf['table']['prefix'] . $conf['table']['data_summary_ad_zone_assoc'];
        $joinTable         = $conf['table']['prefix'] . $conf['table']['data_intermediate_ad'];
        $query['table']    = $table;
        $query['fields']   = array(
                                "$table.ad_id",
                                "$table.zone_id",
                                "$table.impressions_requested",
                                "$table.priority_factor",
                                "$table.past_zone_traffic_fraction",
                                "SUM($joinTable.impressions) AS impressions"
                             );
        $query['wheres']   = array(
                                array("$table.operation_interval = {$conf['maintenance']['operationInterval']}", 'AND'),
                                array("$table.operation_interval_id = $previousOptIntID", 'AND'),
                                array("$table.interval_start = '" . $aDates['start']->format('%Y-%m-%d %H:%M:%S') . "'", 'AND'),
                                array("$table.interval_end = '" . $aDates['end']->format('%Y-%m-%d %H:%M:%S') . "'", 'AND'),
                                array("$table.expired IS NULL", 'AND')
                             );
        $query['joins']    = array(
                                array($joinTable, "$table.ad_id = $joinTable.ad_id AND $table.zone_id = $joinTable.zone_id")
                             );
        $query['group']    = "$table.ad_id, $table.zone_id";
        $query['orderBys'] = array(
                        array("$table.ad_id", 'ASC'),
                        array("$table.zone_id", 'ASC')
                             );
        return $this->_get($query);
    }
    
    /**
     * A method to update the priority values stored in the database.
     *
     * Requires that a current day/time (as a Date object) be registered
     * with the ServiceLocator (as "now").
     *
     * @access public
     * @param array $aData An array of zones, indexed by zone ID, each containing an array
     *                     "ads", which in turn is an array of hashes, each containing the
     *                     following indexes/data:
     *                      - "ad_id"                 The ad ID.
     *                      - "zone_id"               The zone ID.
     *                      - "priority"              The ad/zone priority value.
     *                      - "requested_impressions" The number of impressions the priority
     *                                                should result in.
     *                      - "priority_factor"       The priority adjustment factor used to
     *                                                compensate for filters.
     *                      - "past_zone_traffic_fraction"  The fraction of the zone's impressions
     *                                                      given to the ad in the previous operation
     *                                                      interval.
     * @return boolean True on success, false on failure.
     *
     * @TODO Update the "created_by", "expired_by" fields once Max has a role-based
     *       permissions system in place.
     * @TODO Write the code that inserts non-static priorities.
     */
    function updatePriorities(&$aData)
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $oServiceLocator = &ServiceLocator::instance();
        $oDate = &$oServiceLocator->get('now');
        if (!$oDate) {
            return false;
        }
        $oOperationInterval = &OperationInterval::singleton();
        $currentOperationIntervalID = $oOperationInterval->getIntervalID($oDate);
        $aDates = $oOperationInterval->convertDateToOperationIntervalStartAndEndDates($oDate);
        // Start a transaction
        $this->dbh->startTransaction();
        // Delete all category-based (ie. link_type = MAX_AD_ZONE_LINK_CATEGORY) priorities
        // from ad_zone_assoc
        $query = "
            DELETE FROM
                {$conf['table']['prefix']}{$conf['table']['ad_zone_assoc']}
            WHERE
                link_type = " . MAX_AD_ZONE_LINK_CATEGORY;
        $result = $this->dbh->query($query);
        if (PEAR::isError($result)) {
            $this->dbh->rollback();
            return false;
        }
        // Set all remaining normal (ie. link_type = MAX_AD_ZONE_LINK_NORMAL) priorities to zero
        $query = "
            UPDATE
                {$conf['table']['prefix']}{$conf['table']['ad_zone_assoc']}
            SET
                priority = 0
            WHERE
                link_type = " . MAX_AD_ZONE_LINK_NORMAL;
        $result = $this->dbh->query($query);
        if (PEAR::isError($result)) {
            $this->dbh->rollback();
            return false;
        }
        // Update the static priority values
        if (is_array($aData) && (count($aData) > 0)) {
            foreach ($aData as $aZoneData) {
                if (is_array($aZoneData['ads']) && (count($aZoneData['ads']) > 0)) {
                    foreach ($aZoneData['ads'] as $aAdZonePriority) {
                        $query = "
                            UPDATE
                                {$conf['table']['prefix']}{$conf['table']['ad_zone_assoc']}
                            SET
                                priority = {$aAdZonePriority['priority']}
                            WHERE
                                ad_id = {$aAdZonePriority['ad_id']}
                                AND zone_id = {$aAdZonePriority['zone_id']}";
                        $result = $this->dbh->query($query);
                        if (PEAR::isError($result)) {
                            $this->dbh->rollback();
                            return false;
                        }
                    }
                }
            }
        }
        // Insert the non-static priority values
        
        // Expire the old priority values in data_summary_ad_zone_assoc
        $query = "
            UPDATE
                {$conf['table']['prefix']}{$conf['table']['data_summary_ad_zone_assoc']}
            SET
                expired = '" . $oDate->format('%Y-%m-%d %H:%M:%S') . "',
                expired_by = 0
            WHERE
                expired IS NULL";
        $result = $this->dbh->query($query);
        if (PEAR::isError($result)) {
            $this->dbh->rollback();
            return false;
        }
        // Add the new priority values to data_summary_ad_zone_assoc
        if (is_array($aData) && (count($aData) > 0)) {
            $values = array();
            foreach ($aData as $aZoneData) {
                if (is_array($aZoneData['ads']) && (count($aZoneData['ads']) > 0)) {
                    foreach ($aZoneData['ads'] as $aAdZonePriority) {
                        $values[] = "
                            (
                                {$conf['maintenance']['operationInterval']},
                                $currentOperationIntervalID,
                                '" . $aDates['start']->format('%Y-%m-%d %H:%M:%S') . "',
                                '" . $aDates['end']->format('%Y-%m-%d %H:%M:%S') . "',
                                {$aAdZonePriority['ad_id']},
                                {$aAdZonePriority['zone_id']},
                                {$aAdZonePriority['impressions_requested']},
                                {$aAdZonePriority['priority']},
                                " . (is_null($aAdZonePriority['priority_factor']) ? 'NULL' : $aAdZonePriority['priority_factor']) . ",
                                " . (is_null($aAdZonePriority['past_zone_traffic_fraction']) ? 'NULL' : $aAdZonePriority['past_zone_traffic_fraction']) . ",
                                '" . $oDate->format('%Y-%m-%d %H:%M:%S') . "',
                                0
                            )";
                    }
                }
            }
            if (is_array($values) && (count($values) > 0)) {
                $query = "
                    INSERT INTO
                        {$conf['table']['prefix']}{$conf['table']['data_summary_ad_zone_assoc']}
                        (
                            operation_interval,
                            operation_interval_id,
                            interval_start,
                            interval_end,
                            ad_id,
                            zone_id,
                            impressions_requested,
                            priority,
                            priority_factor,
                            past_zone_traffic_fraction,
                            created,
                            created_by
                        )
                     VALUES" . implode(', ', $values);
                $result = $this->dbh->query($query);
                if (PEAR::isError($result)) {
                    $this->dbh->rollback();
                    return false;
                }
            }
        }
        // Commit the transaction
        $this->dbh->commit();
        return true;
    }
    
    /**
     * A method to obtain the largest active tracker to ad impression 
     * connection window in the database.
     *
     * @return integer The largest found connection window.
     */
    function maxAdImpressionWindow()
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $query = "
            SELECT
                MAX(ct.viewwindow) AS max
            FROM
                {$conf['table']['prefix']}{$conf['table']['campaigns_trackers']} AS ct";
        MAX::debug('Finding the largest active impression connection window for the AdServer module', PEAR_LOG_DEBUG);
        $aRow = $this->dbh->getRow($query);
        if (PEAR::isError($aRow)) {
            return MAX::raiseError($aRow, MAX_ERROR_DBFAILURE);
        }
        return $aRow['max'];
        
    }
    
    /**
     * A method to obtain the largest active tracker to ad click
     * connection window in the database.
     *
     * @return integer The largest found connection window.
     */
    function maxAdClickWindow()
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $query = "
            SELECT
                MAX(ct.clickwindow) AS max
            FROM
                {$conf['table']['prefix']}{$conf['table']['campaigns_trackers']} AS ct";
        MAX::debug('Finding the largest active click connection window for the AdServer module', PEAR_LOG_DEBUG);
        $aRow = $this->dbh->getRow($query);
        if (PEAR::isError($aRow)) {
            return MAX::raiseError($aRow, MAX_ERROR_DBFAILURE);
        }
        return $aRow['max'];
    }
    
    /**
     * A method to get the average number of impressions recorded -- calculated
     * over the required number of past weeks -- for each active zone.
     *
     * @param array $aZoneIds       array of zoneIds to find averages of
     * @param object $oStartDate    Date object representing the beginning of the start 
     *                              operation interval range to be evaluated 
     * @param object $oEndDate      Date object representing the beginning of the end 
     *                              operation interval in range to be evaluated
     * @param integer $weeks        The number of past weeks to average over.
     * @return array
     *
     *  <pre>
     *  Array =>
     *      [zone_id] => Array
     *          (
     *              [operation_interval_id] => Array
     *                  (
     *                      ['average_impressions'] => Average Impressions
     *                  )
     *              [operation_interval_id] => Array
     *                  (
     *                      ['average_impressions'] => Average Impressions
     *                  )
     *          )
     *      [zone_id] => Array
     *          (
     *              [operation_interval_id] => Array
     *                  (
     *                      ['average_impressions'] => Average Impressions
     *                  )
     *              [operation_interval_id] => Array
     *                  (
     *                      ['average_impressions'] => Average Impressions
     *                  )
     *          )
     *  </pre>
     */
     function getZonesImpressionAverageByRange($aZoneIds, $oStartDate, $oEndDate, $weeks)
     {
        if ((!is_array($aZoneIds)) || (!count($aZoneIds))) {
            return;
        }
        $conf = $GLOBALS['_MAX']['CONF'];
        // Clone date parameters
        $oStartDateCopy = &new Date($oStartDate);
        $oEndDateCopy   = &new Date($oEndDate);
        // Get database connection & operation interval
        $dbh = &MAX_DB::singleton();
        $operationInterval = OperationInterval::singleton();
        // Construct the date constraints for the query
        $dateConstraints = "(";
        $aDateConstraints = array();
        for($i = 0; $i < $weeks; $i++) {
            $oStartDateCopy->subtractSeconds(SECONDS_PER_WEEK);
            $oEndDateCopy->subtractSeconds(SECONDS_PER_WEEK);
            $aDateConstraints[] = "
                (
                    interval_start >= '" . $oStartDateCopy->format('%Y-%m-%d %H:%M:%S') . "'
                    AND interval_start <= '" . $oEndDateCopy->format('%Y-%m-%d %H:%M:%S') . "'
                )";
        }
        $dateConstraints .= implode(' OR ', $aDateConstraints);
        $dateConstraints .= ")";
        // Form the database query
        $query = "
            SELECT
                zone_id AS zone_id,
                operation_interval_id AS operation_interval_id,
                ROUND(SUM(actual_impressions)/COUNT(actual_impressions)) AS average_impressions
            FROM
                {$conf['table']['prefix']}{$conf['table']['data_summary_zone_impression_history']}
            WHERE
                (zone_id = " . implode(' OR zone_id = ', $aZoneIds) . ")
                AND operation_interval = $operationInterval->interval
                AND " . $dateConstraints . "
            GROUP BY
                zone_id, operation_interval_id
            HAVING
                COUNT(actual_impressions) = $weeks";
        // Execute the query
        $res = $dbh->query($query);
        // Store the average impressions data
        if (!(PEAR::isError($res))) {
            $return = array();
            while ($res->fetchInto($row)) {
                $return[$row['zone_id']][$row['operation_interval_id']] = $row;
            }
            return $return;
        } else {
            return $res; 
        }
    }
    
    /**
     * A method to get the number of forecast and actual impressions recorded 
     * for each active zone, over a given range.
     *
     * @param array $aZoneIds    array of zoneIds to find averages of
     * @param object $oStartDate Date representing start of operation interval for range
     * @param object $oEndDate   Date representing start of the end operation interval for range
     * @return array
     *  <pre>
     *   Array
     *   (
     *       [zone_id] => Array
     *           (
     *               [operation_interval_id] => Array
     *                   (
     *                       [zone_id] => 1
     *                       [operation_interval_id] => 1
     *                       [forecast_impressions] => 200
     *                       [actual_impressions] => 400
     *                   )
     *   
     *               [operation_interval_id] => Array
     *                   (
     *                       [zone_id] => 1
     *                       [operation_interval_id] => 2
     *                       [forecast_impressions] => 200
     *                       [actual_impressions] => 400
     *                   )
     *           )
     *   
     *   )
     *      .
     *      .
     *      .
     *  </pre>
     */
    function getZonesImpressionHistoryByRange($aZoneIds, $startDate, $endDate)
    {
        if ((!is_array($aZoneIds)) || (!count($aZoneIds))) {
            return;
        }
        $conf = $GLOBALS['_MAX']['CONF'];
        $dbh = &MAX_DB::singleton();
        // Get operation interval
        $operationInterval = OperationInterval::singleton();
        // Construct the SQL to obtain the impressions
        $query = "
            SELECT
                zone_id AS zone_id,
                operation_interval_id AS operation_interval_id,
                forecast_impressions AS forecast_impressions,
                actual_impressions AS actual_impressions
            FROM
                {$conf['table']['prefix']}{$conf['table']['data_summary_zone_impression_history']}
            WHERE
                (zone_id = ";
        $query .= implode(' OR zone_id = ', $aZoneIds);
        $query .= ")
                AND operation_interval = $operationInterval->interval
                AND interval_start >= '" . $startDate->format('%Y-%m-%d %H:%M:%S') . "'
                AND interval_start <= '" . $endDate->format('%Y-%m-%d %H:%M:%S') . "'";
        $res = $dbh->query($query);
        // Store the impression data
        if (!(PEAR::isError($res))) {
            $return = array();
            while ($res->fetchInto($row)) {
                $return[$row['zone_id']][$row['operation_interval_id']] = $row;
            }
            return $return;
        } else {
            return $res;   
        }
    }
    
    /**
     * A method to save zone impression forecasts to the stats_zone_impression_history table.
     *
     * @param array $aForecasts A reference to an array that contains the zone impression
     *                         forecasts, indexed by zone ID, and then operation interval
     *                         ID. Each entry is expected to be an array with the the
     *                         following indicies: "forecast" => The forecast number of
     *                         impressions for the zone/operation interval;
     *                         "interval_start" => The start Date for the operation interval;
     *                         "interval_end" => The end Date for the operation interval; and
     *                         "operation_interval" => The operation interval in use when
     *                         the forecast was made.
     * @param boolean $safe A flag to indicate if it is safe to simply insert the values
     *                      (true), or if each forecast needs to be checked to see if it
     *                      may already exist (false).
     */
    function saveZoneImpressionForecasts($aForecasts, $safe = true)
    {
        // Check the parameter
        if (!is_array($aForecasts) || (is_array($aForecasts) && (count($aForecasts) == 0))) {
            return;
        }
        $conf = $GLOBALS['_MAX']['CONF'];
        $dbh = &MAX_DB::singleton();
        // Get operation interval libraray
        $oOperationInterval = OperationInterval::singleton();
        if ($safe) {
            // Construct the SQL statements to insert the values
            $query = "
                INSERT INTO
                    {$conf['table']['prefix']}{$conf['table']['data_summary_zone_impression_history']}
                    (
                        zone_id,
                        operation_interval,
                        operation_interval_id,
                        interval_start,
                        interval_end,
                        forecast_impressions
                    )
                VALUES";
            foreach ($aForecasts as $zoneId => $aOperationIntervals) {
                foreach ($aOperationIntervals as $id => $aValues) {
                    $aQuery[] = "($zoneId, {$oOperationInterval->interval}, $id, '" .
                        $aValues['interval_start'] . "', '" .
                        $aValues['interval_end'] . "', {$aValues['forecast_impressions']})";
                }
            }
            $query .= ' ' . implode(', ', $aQuery);
            // Insert the new historical zone impressions forecasts
            $result = $dbh->query($query);
        } else {
            // For each forecast in the array
            foreach ($aForecasts as $zoneId => $aOperationIntervals) {
                foreach ($aOperationIntervals as $id => $aValues) {
                    // Is there already a forecast for the zone?
                    $query = "
                        SELECT
                            forecast_impressions AS forecast_impressions
                        FROM
                            {$conf['table']['prefix']}{$conf['table']['data_summary_zone_impression_history']}
                        WHERE
                            zone_id = $zoneId
                            AND operation_interval = {$oOperationInterval->interval}
                            AND operation_interval_id = $id
                            AND interval_start = '{$aValues['interval_start']}'
                            AND interval_end = '{$aValues['interval_end']}'";
                    $result = $dbh->query($query);
                    if ($result->numRows() > 0) {
                        // Update the existing forecast
                        $query = "
                            UPDATE
                                {$conf['table']['prefix']}{$conf['table']['data_summary_zone_impression_history']}
                            SET
                                forecast_impressions = {$aValues['forecast_impressions']}
                            WHERE
                                zone_id = $zoneId
                                AND operation_interval = {$oOperationInterval->interval}
                                AND operation_interval_id = $id
                                AND interval_start = '{$aValues['interval_start']}'
                                AND interval_end = '{$aValues['interval_end']}'";
                        $result = $dbh->query($query);
                    } else {
                        // Insert a new forecast
                        $query = "
                            INSERT INTO
                                {$conf['table']['prefix']}{$conf['table']['data_summary_zone_impression_history']}
                                (
                                    zone_id,
                                    operation_interval,
                                    operation_interval_id,
                                    interval_start,
                                    interval_end,
                                    forecast_impressions
                                )
                            VALUES
                                (
                                    $zoneId,
                                    {$oOperationInterval->interval},
                                    $id,
                                    '{$aValues['interval_start']}',
                                    '{$aValues['interval_end']}',
                                    {$aValues['forecast_impressions']}
                                )";
                        $result = $dbh->query($query);
                    }
                }
            }
        }
    }

    /**
     * A method to get all active zones in the system.
     *
     * An active zone is a zone that is linked to at least one advertisement
     * (via the ad_zone_assoc table), where the advertisement has its "active"
     * field set to true.
     */
    function getActiveZones()
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $query = array();
        $table             = $conf['table']['prefix'] . $conf['table']['zones'];
        $joinTable1        = $conf['table']['prefix'] . $conf['table']['ad_zone_assoc'];
        $joinTable2        = $conf['table']['prefix'] . $conf['table']['banners'];
        $query['table']    = $table;
        $query['fields']   = array(
                                "$table.zoneid",
                                "$table.zonename",
                                "$table.zonetype"
                             );
        $query['joins']    = array(
                                array($joinTable1, "$table.zoneid = $joinTable1.zone_id"),
                                array($joinTable2, "$joinTable1.ad_id = $joinTable2.bannerid AND $joinTable2.active = 't'")
                             );
        return $this->_get($query);
    }
    
    /**
     * A method to get all of the advertisements in a placement.
     *
     * @param integer $id The placement ID.
     * @return array An array of advertisement array, in the following format:
     *
     *      Array
     *      (
     *          [0] => Array
     *              (
     *                  [ad_id] => 1
     *                  [type] => sql
     *                  [weight] => 1
     *                  [active] => t
     *              )
     *      
     *          [1] => Array
     *              (
     *                  [ad_id] => 2
     *                  [type] => sql
     *                  [weight] => 1
     *                  [active] => t
     *              )
     *      
     *      )
     */
    function getAdsByPlacementId($id)
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $query = array();
        $table = $conf['table']['prefix'] . $conf['table']['banners'];
        $query['table']    = $table;        
        $query['fields']   = array(
                                "$table.bannerid AS ad_id",
                                "$table.storagetype AS type",
                                "$table.weight",
                                "$table.active "
                             );
        $query['wheres']   = array(
                                array("$table.campaignid = $id", 'AND'),
                             );
        return $this->_get($query);        
    }

    /**
     * A method to save the total impressions required for each advert
     * into the temporary tmp_ad_required_impression table.
     *
     * @param array $aData An array of array, each containing the
     *                     "ad_id", and its associated "required_impressions".
     */
    function saveRequiredAdImpressions($aData)
    {
        if (is_array($aData) && (count($aData) > 0)) {
            $values = array();
            foreach ($aData as $data) {
                $values[] = "
                    (
                        {$data['ad_id']},
                        {$data['required_impressions']}
                    )";
            }
            $query = "
                INSERT INTO
                    tmp_ad_required_impression
                    (
                        ad_id,
                        required_impressions
                    )
                VALUES" . implode(', ', $values);
            $this->dbh->query($query);
        }
    }

    /**
     * A method to get the required impressions for a list of advertisements
     * from the tmp_ad_required_impression table.
     *
     * @param array $aAdvertID An array of advertisement IDs.
     * @return array An array of required impressions, indexed by advertisement ID.
     */
    function getRequiredAdImpressions($aAdvertID)
    {
        if (is_array($aAdvertID) && (count($aAdvertID) == 0)) {
            return array();
        }
        $conf = $GLOBALS['_MAX']['CONF'];
        $query = "
            SELECT
                ad_id AS ad_id,
                required_impressions AS required_impressions
            FROM
                tmp_ad_required_impression
            WHERE
                ad_id IN (" . implode(', ', $aAdvertID) . ')';
        // Don't use a PEAR_Error handler
        PEAR::pushErrorHandling(null);
        // Execute the query
        $result = $this->dbh->query($query);
        // Resore the PEAR_Error handler
        PEAR::popErrorHandling();
        if (!PEAR::isError($result)) {
            $aResult = array();
            while ($row = $result->fetchRow()) {
                $aResult[$row['ad_id']] = $row['required_impressions'];
            }
            return $aResult;
        } elseif (PEAR::isError($result, DB_ERROR_NOSUCHTABLE)) {
            return array();
        }
        MAX::raiseError($result, MAX_ERROR_DBFAILURE, PEAR_ERROR_DIE);
    }

    /**
     * A method to get current zone impression forecast(s).
     *
     * Requires that a current day/time (as a Date object) be registered
     * with the ServiceLocator (as "now").
     *
     * @return mixed An array, indexed by zone ID, of the current impression
     *               forecasts, or false if the current datetime not registered
     *               with the ServiceLocator.
     */
    function getZoneImpressionForecasts()
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $oServiceLocator = &ServiceLocator::instance();
        $oDate = &$oServiceLocator->get('now');
        if (!$oDate) {
            return false;
        }
        $oOperationInterval = &OperationInterval::singleton();
        // Get the forecast impressions for the current operation interval
        $currentOpIntID = $oOperationInterval->getIntervalID($oDate);
        $aCurrentDates = $oOperationInterval->convertDateToOperationIntervalStartAndEndDates($oDate);
        $query = array();
        $table             = $conf['table']['prefix'] . $conf['table']['data_summary_zone_impression_history'];
        $query['table']    = $table;        
        $query['fields']   = array(
                                "$table.zone_id AS zone_id",
                                "$table.forecast_impressions AS forecast_impressions"
                             );
        $query['wheres']   = array(
                                 array("$table.operation_interval = {$conf['maintenance']['operationInterval']}", 'AND'),
                                 array("$table.operation_interval_id = $currentOpIntID",  'AND'),
                                 array("$table.interval_start = '" . $aCurrentDates['start']->format('%Y-%m-%d %H:%M:%S') . "'", 'AND'),
                                 array("$table.interval_end = '" . $aCurrentDates['end']->format('%Y-%m-%d %H:%M:%S') . "'", 'AND')
                             );
        $result = $this->_get($query);
        $aResult = array();
        foreach ($result as $row) {
            $aResult[$row['zone_id']] = $row['forecast_impressions'];
        }
        return $aResult;
    }

    /**
     * A method to get the ad/zone associations, based on a list of advertisement IDs.
     * Only returns normal ad/zone associations (ie. link_type = MAX_AD_ZONE_LINK_NORMAL),
     * not special direct selection associations, or category associations.
     *
     * @param array $aAdvertID An array of advertisement IDs.
     * @return mixed An array of arrays, indexed by advertisement ID, with each
     *               sub-array containing the zones that the adversisement is linked to.
     */
    function getAdZoneAssociationsByAds($aAdvertID)
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $query = array();
        $table             = $conf['table']['prefix'] . $conf['table']['ad_zone_assoc'];
        $query['table']    = $table;        
        $query['fields']   = array(
                                "$table.ad_id AS ad_id",
                                "$table.zone_id AS zone_id"
                             );
        $query['wheres']   = array(
                                array("$table.ad_id IN (" . implode(', ', $aAdvertID) . ')', 'AND'),
                                array("$table.link_type = " . MAX_AD_ZONE_LINK_NORMAL, 'AND')
                             );
        $result = $this->_get($query);
        $aResult = array();
        foreach ($result as $row) {
            $aResult[$row['ad_id']][] = array('zone_id' => $row['zone_id']);
        }
        return $aResult;
    }
    
    /**
     * A method to save ad/zone impression allocations to the tmp_ad_zone_impression
     * table.
     *
     * @param array An array of arrays, with each sub-array containing the "ad_id",
     *              "zone_id" and "impressions".
     */
    function saveAllocatedImpressions($aData)
    {
        if (is_array($aData) && (count($aData) > 0)) {
            $aValues = array();
            foreach ($aData as $aItem) {
                $aValues[] = "({$aItem['ad_id']}, {$aItem['zone_id']}, {$aItem['impressions']})";
            }
            $query = "
                INSERT INTO
                    tmp_ad_zone_impression
                    (
                        ad_id,
                        zone_id,
                        impressions
                    )
                VALUES ";
            $query .= implode(', ', $aValues);
            $this->dbh->query($query);
        }
    }
    
    /**
     * A method to get the IDs of all of the (active) agencies in the system.
     */
    function getAllActiveAgencyIds()
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $dbh = &MAX_DB::singleton();
        $query = "
            SELECT
                agencyid
            FROM
                {$conf['table']['prefix']}{$conf['table']['agency']}
            WHERE
                active = ".MAX_AGENCY_ACTIVE."
            ORDER BY
                agencyid";
        $result = $dbh->query($query);
        $aResult = array();
        while ($row = $result->fetchRow()) {
            $aResult[] = $row['agencyid'];
        }
        return $aResult;
    }
    
    /**
     * Obtains a database-level lock for the Maintenance Priority Engine process.
     *
     * @return boolean True if lock was obtained, false otherwise.
     */
    function obtainPriorityLock()
    {
        return parent::obtainLock('mpe');
    }

    /**
     * Releases the database-level lock for the Maintenance Priority Engine process.
     *
     * @return mixed True if lock was released, a PEAR Error otherwise.
     */
    function releasePriorityLock()
    {
        return parent::releaseLock('mpe');
    }
    
    
    
    
    

    /**
     * Returns an array of ad limitation hashes.
     *
        Array
        (
            [0] => Array
                (
                    [bannerid] => 3
                    [logical] => and
                    [type] => date
                    [comparison] => !=
                    [data] => 2005-5-25
                    [executionorder] => 0
                )
        
            [1] => Array
         ...
     *
     * @param array $fields
     * @param array $wheres
     * @param array $joins
     * @param array $orderBys
     * @return array
     */
    function getAdLimitations($fields = array(), $wheres = array(), $joins = array(), $orderBys = array())
    {
        return parent::getLimitations($fields, $wheres, $joins, $orderBys, 'acls', 'bannerid');
    }
    
}

?>