<?php

/*
+---------------------------------------------------------------------------+
| Max Media Manager v0.3                                                    |
| =================                                                         |
|                                                                           |
| Copyright (c) 2003-2005 m3 Media Services Limited                         |
| For contact details, see: http://www.m3.net/                              |
|                                                                           |
| This program is free software; you can redistribute it and/or modify      |
| it under the terms of the GNU General Public License as published by      |
| the Free Software Foundation; either version 2 of the License, or         |
| (at your option) any later version.                                       |
|                                                                           |
| This program is distributed in the hope that it will be useful,           |
| but WITHOUT ANY WARRANTY; without even the implied warranty of            |
| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             |
| GNU General Public License for more details.                              |
|                                                                           |
| You should have received a copy of the GNU General Public License         |
| along with this program; if not, write to the Free Software               |
| Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA |
+---------------------------------------------------------------------------+
$Id: mysqlSplit.php 3610 2005-09-01 14:38:04Z andrew $
*/

require_once MAX_PATH . '/lib/Max.php';
require_once MAX_PATH . '/lib/max/Dal/Statistics/AdServer/mysql.php';
require_once 'Date.php';
 
/**
 * The data access layer code for summarising raw data into statistics, for 
 * the AdServer module, when split tables are in use.
 *
 * @package    MaxDal
 * @subpackage Statistics
 * @author     Andrew Hill <andrew@m3.net>
 */
class MAX_Dal_Statistics_AdServer_mysqlSplit extends MAX_Dal_Statistics_AdServer_mysql
{

    /**
     * The constructor method.
     *
     * @uses MAX_Dal_Statistics_AdServer_mysql::MAX_Dal_Statistics_AdServer_mysql()
     */
    function MAX_Dal_Statistics_AdServer_mysqlSplit()
    {
        parent::MAX_Dal_Statistics_AdServer_mysql();
    }
    
    /**
     * A method to find the last time that maintenance statistics was run.
     *
     * @param integer $type The update type that occurred - that is,
     *                      DAL_STATISTICS_COMMON_UPDATE_OI if the update was
     *                      done on the basis of the operation interval, 
     *                      DAL_STATISTICS_COMMON_UPDATE_HOUR if the update
     *                      was done on the basis of the hour, or
     *                      DAL_STATISTICS_COMMON_UPDATE_BOTH if the update
     *                      was done on the basis of both the operation
     *                      interval and the hour.
     * @param Date $now An optional Date, used to specify the "current time", and
     *                  to limit the method to only look for past maintenance
     *                  statistics runs before this date. Normally only used
     *                  to assist with re-generation of statistics in the event
     *                  of faulty raw tables.
     * @return Date A Date representing the date up to which the statistics
     *              have been summarised, for the specified update type, or
     *              the appropriate date based on raw data if maintenance
     *              statistics has never been run for the Max module before.
     *              Returns null if no raw data is available.
     */
    function getMaintenanceStatisticsLastRunInfo($type, $now = null)
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $table = $conf['table']['prefix'] .
                 $conf['table']['data_raw_ad_impression'] . '_' .
                 date('Ymd');
        return $this->_getMaintenanceStatisticsLastRunInfo($type, $table, $now);
    }
    
    
   /**
    * Method to summarise request, impression or click data.
    *
    * @param String $type type of data to summarise
    * @param Date $oStart start of interval to be summarised
    * @param Date $oEnd end of interval to be summarised
    * @return Integer affected rows
    */
    function _summariseData($type, $oStart, $oEnd)
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        if (empty($type)) {
    		return 0;	
    	}
    	$type = strtolower($type);
    	$tmpTableName = 'tmp_ad_' . $type;
    	$countColumnName = $type . 's';
        // Check the start and end dates
        if (!$this->maintenance->checkIntervalDates($oStart, $oEnd, $conf['maintenance']['operationInterval'])) {
            return 0;
        }
        // Get the start and end dates of the operation interval ID
        list($operationIntervalStart, $operationIntervalEnd) =
            $this->maintenance->convertDateToOperationIntervalStartAndEndDates($oStart);
        // How many days does the operation interval span?
        $days = Date_Calc::dateDiff($operationIntervalStart->getDay(),
                                    $operationIntervalStart->getMonth(),
                                    $operationIntervalStart->getYear(),
                                    $operationIntervalEnd->getDay(),
                                    $operationIntervalEnd->getMonth(),
                                    $operationIntervalEnd->getYear());
        // Get the operation interval ID
        $operationIntervalID = $this->maintenance->convertDateToOperationIntervalID($operationIntervalStart);
        // Create the temporary summary table
        $this->tempTables->createTable($tmpTableName);
        // Set the MySQL sort buffer size
        $this->setSortBufferSize();
        // Summarise the data
        $rows = 0;
        $currentDate = &new Date();
        $currentDate->copy($operationIntervalStart);
        for ($counter = 0; $counter <= $days; $counter++) {
            // Set the appropriate data_raw_ad_request table
            $requestTable = $conf['table']['prefix'] . 
                            $conf['table']['data_raw_ad_' . $type] . '_' .
                            $currentDate->format('%Y%m%d');
            $query = "
                INSERT INTO
                    $tmpTableName
                    (
                        day,
                        hour,
                        operation_interval,
                        operation_interval_id,
                        interval_start,
                        interval_end,
                        ad_id,
                        creative_id,
                        zone_id,
                        $countColumnName
                    )
                SELECT
                    DATE_FORMAT(drar.date_time, '%Y-%m-%d') AS day,
                    DATE_FORMAT(drar.date_time, '%k') AS hour,
                    {$conf['maintenance']['operationInterval']} AS operation_interval,
                    $operationIntervalID AS operation_interval_id,
                    '".$operationIntervalStart->format('%Y-%m-%d %H:%M:%S')."' AS interval_start,
                    '".$operationIntervalEnd->format('%Y-%m-%d %H:%M:%S')."' AS interval_end,
                    drar.ad_id AS ad_id,
                    drar.creative_id AS creative_id,
                    drar.zone_id AS zone_id,
                    COUNT(*) AS $countColumnName
                FROM
                    $requestTable AS drar
                WHERE
                    drar.date_time >= ".$oStart->format('%Y%m%d%H%M%S')."
                    AND drar.date_time <= ".$oEnd->format('%Y%m%d%H%M%S')."
                GROUP BY
                    day, hour, ad_id, creative_id, zone_id";
            MAX::debug("Summarising ad $type from the $requestTable table", PEAR_LOG_DEBUG);
            PEAR::pushErrorHandling(null);
            $result = $this->dbh->query($query);
            PEAR::popErrorHandling();
            if (!PEAR::isError($result)) {
                $rows += $this->dbh->affectedRows();
            } elseif (!PEAR::isError($result, DB_ERROR_NOSUCHTABLE)) {
                MAX::raiseError($result, MAX_ERROR_DBFAILURE, PEAR_ERROR_DIE);
            }
            // Update the split table being used
            $currentDate = $currentDate->getNextDay();
        }
        // Restore the MySQL sort buffer size
        $this->restoreSortBufferSize();
        return $rows;
    }
    
    /**
     * A method for summarising requests into a temporary table.
     *
     * @param Date $oStart The start date/time to summarise from.
     * @param Date $oEnd The end date/time to summarise to.
     * @return integer The number of request rows summarised.
     */
    function summariseRequests($oStart, $oEnd)
    {
        return $this->_summariseData('request', $oStart, $oEnd);
    }
    
    /**
     * A method for summarising impressions into a temporary table.
     *
     * @param Date $oStart The start date/time to summarise from.
     * @param Date $oEnd The end date/time to summarise to.
     * @return integer The number of impression rows summarised.
     */
    function summariseImpressions($oStart, $oEnd)
    {
        return $this->_summariseData('impression', $oStart, $oEnd);
    }

    /**
     * A method for summarising clicks into a temporary table.
     *
     * @param Date $oStart The start date/time to summarise from.
     * @param Date $oEnd The end date/time to summarise to.
     * @return integer The number of click rows summarised.
     */    
    function summariseClicks($oStart, $oEnd)
    {
        return $this->_summariseData('click', $oStart, $oEnd);
    }
    
    /**
     * A method for summarising connections into a temporary table.
     *
     * @param Date $start The start date/time to summarise from.
     * @param Date $end The end date/time to summarise to.
     * @return integer The number of connection rows summarised.
     */
    function summariseConnections($start, $end)
    { 
        $conf = $GLOBALS['_MAX']['CONF'];
        if (!$conf['modules']['Tracker']) {
            return 0;
        }
        // Check the start and end dates
        if (!$this->maintenance->checkIntervalDates($start, $end, $conf['maintenance']['operationInterval'])) {
            return 0;
        }
        // Get the start and end dates of the operation interval ID
        list($operationIntervalStart, $operationIntervalEnd) =
            $this->maintenance->convertDateToOperationIntervalStartAndEndDates($start);
        // How many days does the operation interval span?
        $days = Date_Calc::dateDiff($operationIntervalStart->getDay(),
                                    $operationIntervalStart->getMonth(),
                                    $operationIntervalStart->getYear(),
                                    $operationIntervalEnd->getDay(),
                                    $operationIntervalEnd->getMonth(),
                                    $operationIntervalEnd->getYear());
        // Get the operation interval ID
        $operationIntervalID = $this->maintenance->convertDateToOperationIntervalID($operationIntervalStart);
        // Get the maximum impression and click connection windows
        list($impressionWindow, $clickWindow) = $this->maintenance->maxConnectionWindows();
        // Create the tmp_ad_connection table
        $this->tempTables->createTable('tmp_ad_connection');       
        // Initialise rows
		$rows = 0;
        // Insert tracker impressions that may connect to ad impressions into
        // the temporary tmp_tracker_impression_ad_impression_connection table
        $currentDate = &new Date();
        $currentDate->copy($operationIntervalStart);
        for ($counter = 0; $counter <= $days; $counter++) {
            // Create the tmp_tracker_impression_ad_impression_connection table
            $this->tempTables->createTable('tmp_tracker_impression_ad_impression_connection');
            // Set the appropriate data_raw_tracker_impression table
            $trackerImpressionTable = $conf['table']['prefix'] . 
                                      $conf['table']['data_raw_tracker_impression'] . '_' .
                                      $currentDate->format('%Y%m%d');
            $query = "
                INSERT INTO
                    tmp_tracker_impression_ad_impression_connection
                    (
                        server_raw_tracker_impression_id,
                        server_raw_ip,
                        tracker_id,
                        campaign_id,
                        operation_interval,
                        operation_interval_id,
                        interval_start,
                        interval_end,
                        viewer_id,
                        date_time,
                        connection_type,
                        connection_window,
                        connection_status
                    )
                SELECT
                    drti.server_raw_tracker_impression_id AS server_raw_tracker_impression_id,
                    drti.server_raw_ip AS server_raw_ip,
                    drti.tracker_id AS tracker_id,
                    ct.campaignid AS campaign_id,
                    {$conf['maintenance']['operationInterval']} AS operation_interval,
                    $operationIntervalID AS operation_interval_id,
                    '".$operationIntervalStart->format('%Y-%m-%d %H:%M:%S')."' AS interval_start,
                    '".$operationIntervalEnd->format('%Y-%m-%d %H:%M:%S')."' AS interval_end,
                    drti.viewer_id AS viewer_id,
                    drti.date_time AS date_time,
                    '" . MAX_CONNECTION_AD_IMPRESSION . "' AS connection_type,
                    ct.viewwindow AS connection_window,
                    ct.status AS connection_status
                FROM
                    $trackerImpressionTable AS drti,        
                    {$conf['table']['prefix']}{$conf['table']['campaigns_trackers']} AS ct
                WHERE        
                    drti.tracker_id = ct.trackerid
                    AND drti.viewer_id != ''
                    AND drti.date_time >= ".$start->format('%Y%m%d%H%M%S')."
                    AND drti.date_time <= ".$end->format('%Y%m%d%H%M%S')."
                    AND ct.viewwindow > 0";
            MAX::debug('Selecting tracker impressions that may connect to ad impressions into the ' .
                       'tmp_tracker_impression_ad_impression_connection temporary table', PEAR_LOG_DEBUG);
            PEAR::pushErrorHandling(null);
            $result = $this->dbh->query($query);
            PEAR::popErrorHandling();
            if (!PEAR::isError($result)) {
                if ($this->dbh->affectedRows() > 0) {
                    // Connect the tracker impressions with ad impressions, where possible, by looking
                    // over the necessary past data_raw_ad_impression tables, as appropriate for the
                    // maximum impression connection window
                    $endImpressionDate = $currentDate;
                    $startImpressionDate = &new Date();
                    $startImpressionDate->copy($endImpressionDate);
                    $startImpressionDate->subtractSeconds((int) $impressionWindow); // Cast to int as Date class
                                                                                    // doesn't correctly deal with
                                                                                    // integers as strings...
                    // How many days does the impression connection window span?
                    $impressionDays = Date_Calc::dateDiff($startImpressionDate->getDay(),
                                                          $startImpressionDate->getMonth(),
                                                          $startImpressionDate->getYear(),
                                                          $endImpressionDate->getDay(),
                                                          $endImpressionDate->getMonth(),
                                                          $endImpressionDate->getYear());
                    $currentImpressionDate = &new Date();
                    $currentImpressionDate->copy($startImpressionDate);
                    for ($impressionCounter = 0; $impressionCounter <= $impressionDays; $impressionCounter++) {
                        // Set the appropriate data_raw_ad_impression table
                        $adImpressionTable = $conf['table']['prefix'] . 
                                             $conf['table']['data_raw_ad_impression'] . '_' .
                                             $currentImpressionDate->format('%Y%m%d');
                        $query = "
                            INSERT INTO
                                tmp_ad_connection
                                (
                                    server_raw_tracker_impression_id,
                                    server_raw_ip,
                                    date_time,
                                    operation_interval,
                                    operation_interval_id,
                                    interval_start,
                                    interval_end,
                                    connection_viewer_id,
                                    connection_viewer_session_id,
                                    connection_date_time,
                                    connection_ad_id,
                                    connection_creative_id,
                                    connection_zone_id,
                                    connection_channel,
                                    connection_language,
                                    connection_ip_address,
                                    connection_host_name,
                                    connection_country,
                                    connection_https,
                                    connection_domain,
                                    connection_page,
                                    connection_query,
                                    connection_referer,
                                    connection_search_term,
                                    connection_user_agent,
                                    connection_os,
                                    connection_browser,
                                    connection_type,
                                    connection_window,
                                    connection_status
                                )
                            SELECT DISTINCT
                                ttiaic.server_raw_tracker_impression_id AS server_raw_tracker_impression_id,
                                ttiaic.server_raw_ip AS server_raw_ip,
                                ttiaic.date_time AS date_time,
                                ttiaic.operation_interval AS operation_interval,
                                ttiaic.operation_interval_id AS operation_interval_id,
                                ttiaic.interval_start AS interval_start,
                                ttiaic.interval_end AS interval_end,
                                drai.viewer_id AS connection_viewer_id,
                                drai.viewer_session_id AS connection_viewer_session_id,
                                drai.date_time AS connection_date_time,
                                drai.ad_id AS connection_ad_id,
                                drai.creative_id AS connection_creative_id,
                                drai.zone_id AS connection_zone_id,
                                drai.channel AS connection_channel,
                                drai.language AS connection_language,
                                drai.ip_address AS connection_ip_address,
                                drai.host_name AS connection_host_name,
                                drai.country AS connection_country,
                                drai.https AS connection_https,
                                drai.domain AS connection_domain,
                                drai.page AS connection_page,
                                drai.query AS connection_query,
                                drai.referer AS connection_referer,
                                drai.search_term AS connection_search_term,
                                drai.user_agent AS connection_user_agent,
                                drai.os AS connection_os,
                                drai.browser AS connection_browser,
                                ttiaic.connection_type AS connection_type,
                                ttiaic.connection_window AS connection_window,
                                ttiaic.connection_status AS connection_status
                            FROM
                                tmp_tracker_impression_ad_impression_connection AS ttiaic,
                                $adImpressionTable AS drai,
                                {$conf['table']['prefix']}{$conf['table']['banners']} AS b,
                                {$conf['table']['prefix']}{$conf['table']['campaigns_trackers']} AS ct
                            WHERE
                                ttiaic.viewer_id = drai.viewer_id
                                AND drai.ad_id = b.bannerid
                                AND b.campaignid = ct.campaignid
                                AND ct.trackerid = ttiaic.tracker_id
                                AND ttiaic.campaign_id = ct.campaignid
                                AND ttiaic.date_time < DATE_ADD(drai.date_time, INTERVAL ttiaic.connection_window SECOND)
                                AND ttiaic.date_time >= drai.date_time";
                        MAX::debug("Connecting tracker impressions with ad impressions from the $adImpressionTable " .
                                   'table, where possible', PEAR_LOG_DEBUG);
                        PEAR::pushErrorHandling(null);
                        $result = $this->dbh->query($query);
                        PEAR::popErrorHandling();
                        if (!PEAR::isError($result)) {
                            $rows += $this->dbh->affectedRows();
                        } elseif (!PEAR::isError($result, DB_ERROR_NOSUCHTABLE)) {
                            MAX::raiseError($result, MAX_ERROR_DBFAILURE, PEAR_ERROR_DIE);
                        }
                        // Update the split data_raw_ad_impression table being used
                        $currentImpressionDate = $currentImpressionDate->getNextDay();
                    }
                }
            } elseif (!PEAR::isError($result, DB_ERROR_NOSUCHTABLE)) {
                MAX::raiseError($result, MAX_ERROR_DBFAILURE, PEAR_ERROR_DIE);
            }
            // Drop the temporary table for storing tracker impressions that may connect to ad impressions
            $this->tempTables->dropTempTable('tmp_tracker_impression_ad_impression_connection');
            // Update the split data_raw_ad_impression table being used
            $currentDate = $currentDate->getNextDay();
        } 
        // Insert tracker impressions that may connect to ad clicks into
        // the temporary tmp_tracker_impression_ad_click_connection table
        $currentDate = &new Date();
        $currentDate->copy($operationIntervalStart);
        for ($counter = 0; $counter <= $days; $counter++) {
            // Create the tmp_tracker_impression_ad_click_connection table
            $this->tempTables->createTable('tmp_tracker_impression_ad_click_connection');
            // Set the appropriate data_raw_tracker_impression table
            $trackerImpressionTable = $conf['table']['prefix'] . 
                                      $conf['table']['data_raw_tracker_impression'] . '_' .
                                      $currentDate->format('%Y%m%d');
            $query = "
                INSERT INTO
                    tmp_tracker_impression_ad_click_connection
                    (
                        server_raw_tracker_impression_id,
                        server_raw_ip,
                        tracker_id,
                        campaign_id,
                        operation_interval,
                        operation_interval_id,
                        interval_start,
                        interval_end,
                        viewer_id,
                        date_time,
                        connection_type,
                        connection_window,
                        connection_status
                    )
                SELECT
                    drti.server_raw_tracker_impression_id AS server_raw_tracker_impression_id,
                    drti.server_raw_ip AS server_raw_ip,
                    drti.tracker_id AS tracker_id,
                    ct.campaignid AS campaign_id,
                    {$conf['maintenance']['operationInterval']} AS operation_interval,
                    $operationIntervalID AS operation_interval_id,
                    '".$operationIntervalStart->format('%Y-%m-%d %H:%M:%S')."' AS interval_start,
                    '".$operationIntervalEnd->format('%Y-%m-%d %H:%M:%S')."' AS interval_end,
                    drti.viewer_id AS viewer_id,
                    drti.date_time AS date_time,
                    '" . MAX_CONNECTION_AD_CLICK . "' AS connection_type,
                    ct.clickwindow AS connection_window,
                    ct.status AS connection_status
                FROM
                    $trackerImpressionTable AS drti,        
                    {$conf['table']['prefix']}{$conf['table']['campaigns_trackers']} AS ct
                WHERE        
                    drti.tracker_id = ct.trackerid
                    AND drti.viewer_id != ''
                    AND drti.date_time >= ".$start->format('%Y%m%d%H%M%S')."
                    AND drti.date_time <= ".$end->format('%Y%m%d%H%M%S')."
                    AND ct.clickwindow > 0";
            PEAR::pushErrorHandling(null);
            $result = $this->dbh->query($query);
            PEAR::popErrorHandling();
            if (!PEAR::isError($result)) {
                if ($this->dbh->affectedRows() > 0) {
                    // Connect the tracker impressions with ad clicks, where possible, by looking
                    // over the necessary past data_raw_ad_click tables, as appropriate for the
                    // maximum click connection window
                    $endClickDate = $currentDate;
                    $startClickDate = &new Date();
                    $startClickDate->copy($endClickDate);
                    $startClickDate->subtractSeconds((int) $clickWindow); // Cast to int as Date class
                                                                          // doesn't correctly deal with
                                                                          // integers as strings...
                    // How many days does the impression connection window span?
                    $clickDays = Date_Calc::dateDiff($startClickDate->getDay(),
                                                     $startClickDate->getMonth(),
                                                     $startClickDate->getYear(),
                                                     $endClickDate->getDay(),
                                                     $endClickDate->getMonth(),
                                                     $endClickDate->getYear());
                    $currentClickDate = &new Date();
                    $currentClickDate->copy($startClickDate);
                    for ($clickCounter = 0; $clickCounter <= $clickDays; $clickCounter++) {
                        // Set the appropriate data_raw_ad_click table
                        $adClickTable = $conf['table']['prefix'] . 
                                        $conf['table']['data_raw_ad_click'] . '_' .
                                        $currentClickDate->format('%Y%m%d');
                        $query = "
                            INSERT INTO
                                tmp_ad_connection
                                (
                                    server_raw_tracker_impression_id,
                                    server_raw_ip,
                                    date_time,
                                    operation_interval,
                                    operation_interval_id,
                                    interval_start,
                                    interval_end,
                                    connection_viewer_id,
                                    connection_viewer_session_id,
                                    connection_date_time,
                                    connection_ad_id,
                                    connection_creative_id,
                                    connection_zone_id,
                                    connection_channel,
                                    connection_language,
                                    connection_ip_address,
                                    connection_host_name,
                                    connection_country,
                                    connection_https,
                                    connection_domain,
                                    connection_page,
                                    connection_query,
                                    connection_referer,
                                    connection_search_term,
                                    connection_user_agent,
                                    connection_os,
                                    connection_browser,
                                    connection_type,
                                    connection_window,
                                    connection_status
                                )
                            SELECT DISTINCT
                                ttiacc.server_raw_tracker_impression_id AS server_raw_tracker_impression_id,
                                ttiacc.server_raw_ip AS server_raw_ip,
                                ttiacc.date_time AS date_time,
                                ttiacc.operation_interval AS operation_interval,
                                ttiacc.operation_interval_id AS operation_interval_id,
                                ttiacc.interval_start AS interval_start,
                                ttiacc.interval_end AS interval_end,
                                drac.viewer_id AS connection_viewer_id,
                                drac.viewer_session_id AS connection_viewer_session_id,
                                drac.date_time AS connection_date_time,
                                drac.ad_id AS connection_ad_id,
                                drac.creative_id AS connection_creative_id,
                                drac.zone_id AS connection_zone_id,
                                drac.channel AS connection_channel,
                                drac.language AS connection_language,
                                drac.ip_address AS connection_ip_address,
                                drac.host_name AS connection_host_name,
                                drac.country AS connection_country,
                                drac.https AS connection_https,
                                drac.domain AS connection_domain,
                                drac.page AS connection_page,
                                drac.query AS connection_query,
                                drac.referer AS connection_referer,
                                drac.search_term AS connection_search_term,
                                drac.user_agent AS connection_user_agent,
                                drac.os AS connection_os,
                                drac.browser AS connection_browser,
                                ttiacc.connection_type AS connection_type,
                                ttiacc.connection_window AS connection_window,
                                ttiacc.connection_status AS connection_status
                            FROM
                                tmp_tracker_impression_ad_click_connection AS ttiacc,
                                $adClickTable AS drac,
                                {$conf['table']['prefix']}{$conf['table']['banners']} AS b,
                                {$conf['table']['prefix']}{$conf['table']['campaigns_trackers']} AS ct
                            WHERE
                                ttiacc.viewer_id = drac.viewer_id
                                AND drac.ad_id = b.bannerid
                                AND b.campaignid = ct.campaignid
                                AND ct.trackerid = ttiacc.tracker_id
                                AND ttiacc.campaign_id = ct.campaignid
                                AND ttiacc.date_time < DATE_ADD(drac.date_time, INTERVAL ttiacc.connection_window SECOND)
                                AND ttiacc.date_time >= drac.date_time";
                        MAX::debug("Connecting tracker impressions with ad clicks from the $adClickTable " .
                                   'table, where possible', PEAR_LOG_DEBUG);
                        PEAR::pushErrorHandling(null);
                        $result = $this->dbh->query($query);
                        PEAR::popErrorHandling();
                        if (!PEAR::isError($result)) {
                            $rows += $this->dbh->affectedRows();
                        } elseif (!PEAR::isError($result, DB_ERROR_NOSUCHTABLE)) {
                            MAX::raiseError($result, MAX_ERROR_DBFAILURE, PEAR_ERROR_DIE);
                        }
                        // Update the split data_raw_ad_click table being used
                        $currentClickDate = $currentClickDate->getNextDay();
                    }
                }
            } elseif (!PEAR::isError($result, DB_ERROR_NOSUCHTABLE)) {
                MAX::raiseError($result, MAX_ERROR_DBFAILURE, PEAR_ERROR_DIE);
            }
            // Drop the temporary table for storing tracker impressions that may connect to ad clicks
            $this->tempTables->dropTempTable('tmp_tracker_impression_ad_click_connection');
            // Update the split data_raw_ad_impression table being used
            $currentDate = $currentDate->getNextDay();
        }
        return $rows;
    }
    
    /**
     * A private method to save the tacker impression/ad impression and
     * tracker impression/ad click connections that have been marked as
     * "latest", and also save the associated tracker impression 
     * variable values.
     *
     * @access private
     * @param Date $start The start date/time to save from.
     * @param Date $end The end date/time to save to.
     */
    function _saveConnectionsAndVariableValues($start, $end)
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        // How many days does the operation interval span?
        $days = Date_Calc::dateDiff($start->getDay(), $start->getMonth(), $start->getYear(),
                                    $end->getDay(),   $end->getMonth(),   $end->getYear());
        $outerTable = $conf['table']['prefix'] .
                      $conf['table']['data_intermediate_ad_connection'];
        $innerTable = $conf['table']['prefix'] .
                       $conf['table']['data_intermediate_ad_variable_value'];
        // Save the connections
        $outerDate = &new Date();
        $outerDate->copy($start);
        for ($counter = 0; $counter <= $days; $counter++) {
            // Set the appropriate data_raw_tracker_impression table
            $trackerImpressionTable = $conf['table']['prefix'] . 
                                      $conf['table']['data_raw_tracker_impression'] . '_' .
                                      $outerDate->format('%Y%m%d');
            $query = "
                INSERT INTO
                    $outerTable
                    (
                        server_raw_tracker_impression_id,
                        server_raw_ip,
                        viewer_id,
                        viewer_session_id,
                        tracker_date_time,
                        connection_date_time,
                        tracker_id,
                        ad_id,
                        creative_id,
                        zone_id,
                        tracker_channel,
                        connection_channel,
                        tracker_language,
                        connection_language,
                        tracker_ip_address,
                        connection_ip_address,
                        tracker_host_name,
                        connection_host_name,
                        tracker_country,
                        connection_country,
                        tracker_https,
                        connection_https,
                        tracker_domain,
                        connection_domain,
                        tracker_page,
                        connection_page,
                        tracker_query,
                        connection_query,
                        tracker_referer,
                        connection_referer,
                        tracker_search_term,
                        connection_search_term,
                        tracker_user_agent,
                        connection_user_agent,
                        tracker_os,
                        connection_os,
                        tracker_browser,
                        connection_browser,
                        connection_type,
                        connection_window,
                        connection_status
                    )
                SELECT
                    drti.server_raw_tracker_impression_id AS server_raw_tracker_impression_id,
                    drti.server_raw_ip AS server_raw_ip,
                    drti.viewer_id AS viewer_id,
                    drti.viewer_session_id AS viewer_session_id,
                    drti.date_time AS tracker_date_time,
                    tac.connection_date_time AS connection_date_time,
                    drti.tracker_id AS tracker_id,
                    tac.connection_ad_id AS ad_id,
                    tac.connection_creative_id AS creative_id,
                    tac.connection_zone_id AS zone_id,
                    drti.channel AS tracker_channel,
                    tac.connection_channel AS connection_channel,
                    drti.language AS tracker_language,
                    tac.connection_language AS connection_language,
                    drti.ip_address AS tracker_ip_address,
                    tac.connection_ip_address AS connection_ip_address,
                    drti.host_name AS tracker_host_name,
                    tac.connection_host_name AS connection_host_name,
                    drti.country AS tracker_country,
                    tac.connection_country AS connection_country,
                    drti.https AS tracker_https,
                    tac.connection_https AS connection_https,
                    drti.domain AS tracker_domain,
                    tac.connection_domain AS connection_domain,
                    drti.page AS tracker_page,
                    tac.connection_page AS connection_page,
                    drti.query AS tracker_query,
                    tac.connection_query AS connection_query,
                    drti.referer AS tracker_referer,
                    tac.connection_referer AS connection_referer,
                    drti.search_term AS tracker_search_term,
                    tac.connection_search_term AS connection_search_term,
                    drti.user_agent AS tracker_user_agent,
                    tac.connection_user_agent AS connection_user_agent,
                    drti.os AS tracker_os,
                    tac.connection_os AS connection_os,
                    drti.browser AS tracker_browser,
                    tac.connection_browser AS connection_browser,
                    tac.connection_type AS connection_type,
                    tac.connection_window AS connection_window,
                    tac.connection_status AS connection_status
                FROM
                    tmp_ad_connection AS tac,
                    $trackerImpressionTable AS drti
                WHERE
                    tac.server_raw_tracker_impression_id = drti.server_raw_tracker_impression_id
                    AND tac.server_raw_ip = drti.server_raw_ip
                    AND tac.latest = 1";
            MAX::debug("Inserting the connections into the $outerTable table", PEAR_LOG_DEBUG);
            PEAR::pushErrorHandling(null);
            $result = $this->dbh->query($query);
            PEAR::popErrorHandling();
            if (PEAR::isError($result) && (!PEAR::isError($result, DB_ERROR_NOSUCHTABLE))) {
                MAX::raiseError($result, MAX_ERROR_DBFAILURE, PEAR_ERROR_DIE);
            }
            // Save any variable values associated with the above connections
            $innerDate = &new Date();
            $innerDate->copy($outerDate);
            for ($innerCounter = 0; $innterCounter < 2; $innterCounter++) {
                // Set the appropriate data_raw_tracker_variable_value table
                $trackerVariableValueTable = $conf['table']['prefix'] . 
                                             $conf['table']['data_raw_tracker_variable_value'] . '_' .
                                             $innerDate->format('%Y%m%d');
                $query = "
                    INSERT INTO
                        $innerTable
                        (
                            data_intermediate_ad_connection_id,
                            tracker_variable_id,
                            value
                        )
                    SELECT
                        diac.data_intermediate_ad_connection_id AS data_intermediate_ad_connection_id,        
                        drtvv.tracker_variable_id AS tracker_variable_id,        
                        drtvv.value AS value
                    FROM
                        {$conf['table']['prefix']}{$conf['table']['data_intermediate_ad_connection']} AS diac,
                        $trackerVariableValueTable AS drtvv,
                        $trackerImpressionTable AS drti
                    WHERE
                        diac.server_raw_tracker_impression_id = drti.server_raw_tracker_impression_id
                        AND diac.server_raw_ip = drti.server_raw_ip
                        AND drti.server_raw_tracker_impression_id = drtvv.server_raw_tracker_impression_id
                        AND drti.server_raw_ip = drtvv.server_raw_ip
                        AND diac.tracker_date_time >= '" . $start->format('%Y-%m-%d %H:%M:%S') . "'
                        AND diac.tracker_date_time <= '" . $end->format('%Y-%m-%d %H:%M:%S') . "'";
                MAX::debug("Saving the tracker impression variable values from the $innerTable table",
                           PEAR_LOG_DEBUG);
                PEAR::pushErrorHandling(null);
                $result = $this->dbh->query($query);
                PEAR::popErrorHandling();
                if (PEAR::isError($result) && (!PEAR::isError($result, DB_ERROR_NOSUCHTABLE))) {
                    MAX::raiseError($result, MAX_ERROR_DBFAILURE, PEAR_ERROR_DIE);
                }
                // Update the split data_raw_tracker_variable_value table being used
                $innerDate = $innerDate->getNextDay();
            }
            // Update the split data_raw_tracker_impression table being used
            $outerDate = $outerDate->getNextDay();
        }
    }
    
    /**
     * A method to delete old (ie. summarised) raw data.
     *
     * @param Date $summarisedTo The date/time up to which data have been summarised (i.e. if there
     *                           is no newer data than this date (minus any compact_stats_grace window)
     *                           then the raw table will be dropped, unless required by the tracking
     *                           module, where installed).
     * @return integer The number of raw tables dropped.
     */
    function deleteOldData($summarisedTo)
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        $now = new Date();  // Current day
        $deleteDate = $summarisedTo;
        if ($conf['maintenance']['compactStatsGrace'] > 0) {
            $deleteDate->subtractSeconds((int) $conf['maintenance']['compactStatsGrace']);
        }
        $tablesDropped = 0;
        // As split tables are in use, look over all possible tables
        $tables = $this->dbh->getListOf('tables');
        if (PEAR::isError($result)) {
            return MAX::raiseError($result, MAX_ERROR_DBFAILURE, PEAR_ERROR_DIE);
        }
        // Delete the requests before taking into account the maximum connection window
        foreach ($tables as $table) {
            // Look at the data_raw_ad_request tables
            if (preg_match('/^' . $conf['table']['prefix'] . $conf['table']['data_raw_ad_request'] . '_' .
                           '(\d{4})(\d{2})(\d{2})$/', $table, $matches)) {
                $date = new Date("{$matches[1]}-{$matches[2]}-{$matches[3]}");
                // Is this today's table?
                if (($date->getYear() == $now->getYear()) &&
                    ($date->getMonth() == $now->getMonth()) &&
                    ($date->getDay() == $now->getDay())) {
                    // Don't drop today's table
                    continue;
                }
                $table = $conf['table']['prefix'] . 
                         $conf['table']['data_raw_ad_request'] . '_' .
                         $date->format('%Y%m%d');
                $query = "
                    SELECT
                        *
                    FROM
                        $table
                    WHERE
                        date_time > '" . $deleteDate->format('%Y-%m-%d %H:%M:%S') ."'
                    LIMIT 1";
                MAX::debug("Selecting non-summarised (later than '" . $deleteDate->format('%Y-%m-%d %H:%M:%S') .
                           "') ad requests from the $table table", PEAR_LOG_DEBUG);
                $innerResult = $this->dbh->query($query);
                if (PEAR::isError($innerResult)) {
                    MAX::raiseError($innerResult, MAX_ERROR_DBFAILURE, PEAR_ERROR_DIE);
                }
                if ($innerResult->numRows() == 0) {
                    // No current data in the table - drop it
                    $query = "DROP TABLE $table";
                    MAX::debug("No non-summarised ad requests in the $table table, so dropping",
                               PEAR_LOG_DEBUG);
                    $innerResult = $this->dbh->query($query);
                    $tablesDropped++;
                }
            }
        }
        // Take into account the maximum connection window, if approproate
        if ($conf['modules']['Tracker']) {
            // Find the largest, active impression and click connection windows
            list($impressionWindow, $clickWindow) = $this->maintenance->maxConnectionWindows();
            // Find the largest of the two windows
            if ($impressionWindow > $clickWindow) {
                $maxWindow = $impressionWindow;
            } else {
                $maxWindow = $clickWindow;
            }
            MAX::debug('Found maximum connection window of ' . $maxWindow . ' seconds', PEAR_LOG_DEBUG);
            $deleteDate->subtractSeconds((int) $maxWindow); // Cast to int, as Date class
                                                            // doesn't deal with strings
        }
        // Delete from remaining tables
        foreach ($tables as $table) {
            // Look at the data_raw_ad_impression tables
            if (preg_match('/^' . $conf['table']['prefix'] . $conf['table']['data_raw_ad_impression'] . '_' .
                           '(\d{4})(\d{2})(\d{2})$/', $table, $matches)) {
                $date = new Date("{$matches[1]}-{$matches[2]}-{$matches[3]}");
                // Is this today's table?
                if (($date->getYear() == $now->getYear()) &&
                    ($date->getMonth() == $now->getMonth()) &&
                    ($date->getDay() == $now->getDay())) {
                    // Don't drop today's table
                    continue;
                }
                $table = $conf['table']['prefix'] . 
                         $conf['table']['data_raw_ad_impression'] . '_' .
                         $date->format('%Y%m%d');
                $query = "
                    SELECT
                        *
                    FROM
                        $table
                    WHERE
                        date_time > '" . $deleteDate->format('%Y-%m-%d %H:%M:%S') ."'
                    LIMIT 1";
                MAX::debug("Selecting non-summarised (later than '" . $deleteDate->format('%Y-%m-%d %H:%M:%S') .
                           "') ad impressions from the $table table", PEAR_LOG_DEBUG);
                $innerResult = $this->dbh->query($query);
                if (PEAR::isError($innerResult)) {
                    MAX::raiseError($innerResult, MAX_ERROR_DBFAILURE, PEAR_ERROR_DIE);
                }
                if ($innerResult->numRows() == 0) {
                    // No current data in the table - drop it
                    $query = "DROP TABLE $table";
                    MAX::debug("No non-summarised ad impressions in the $table table, so dropping",
                               PEAR_LOG_DEBUG);
                    $innerResult = $this->dbh->query($query);
                    $tablesDropped++;
                }
            }
            // Look at the data_raw_ad_click tables
            if (preg_match('/^' . $conf['table']['prefix'] . $conf['table']['data_raw_ad_click'] . '_' .
                           '(\d{4})(\d{2})(\d{2})$/', $table, $matches)) {
                $date = new Date("{$matches[1]}-{$matches[2]}-{$matches[3]}");
                // Is this today's table?
                if (($date->getYear() == $now->getYear()) &&
                    ($date->getMonth() == $now->getMonth()) &&
                    ($date->getDay() == $now->getDay())) {
                    // Don't drop today's table
                    continue;
                }
                $table = $conf['table']['prefix'] . 
                         $conf['table']['data_raw_ad_click'] . '_' .
                         $date->format('%Y%m%d');
                $query = "
                    SELECT
                        *
                    FROM
                        $table
                    WHERE
                        date_time > '" . $deleteDate->format('%Y-%m-%d %H:%M:%S') ."'
                    LIMIT 1";
                MAX::debug("Selecting non-summarised (later than '" . $deleteDate->format('%Y-%m-%d %H:%M:%S') .
                           "') ad clicks from the $table table", PEAR_LOG_DEBUG);
                $innerResult = $this->dbh->query($query);
                if (PEAR::isError($innerResult)) {
                    MAX::raiseError($innerResult, MAX_ERROR_DBFAILURE, PEAR_ERROR_DIE);
                }
                if ($innerResult->numRows() == 0) {
                    // No current data in the table - drop it
                    $query = "DROP TABLE $table";
                    MAX::debug("No non-summarised ad clicks in the $table table, so dropping",
                               PEAR_LOG_DEBUG);
                    $innerResult = $this->dbh->query($query);
                    $tablesDropped++;
                }
            }
        }
        return $tablesDropped;
    }
    
}

?>
