<?php

/*
+---------------------------------------------------------------------------+
| Max Media Manager v0.3                                                    |
| =================                                                         |
|                                                                           |
| Copyright (c) 2003-2005 m3 Media Services Limited                         |
| For contact details, see: http://www.m3.net/                              |
|                                                                           |
| This program is free software; you can redistribute it and/or modify      |
| it under the terms of the GNU General Public License as published by      |
| the Free Software Foundation; either version 2 of the License, or         |
| (at your option) any later version.                                       |
|                                                                           |
| This program is distributed in the hope that it will be useful,           |
| but WITHOUT ANY WARRANTY; without even the implied warranty of            |
| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             |
| GNU General Public License for more details.                              |
|                                                                           |
| You should have received a copy of the GNU General Public License         |
| along with this program; if not, write to the Free Software               |
| Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA |
+---------------------------------------------------------------------------+
$Id: lib-log.inc.php 3659 2005-09-12 10:25:23Z radek $
*/

/**
 * @package    MaxDelivery
 * @author     Andrew Hill <andrew@m3.net>
 * @author     Chris Nutting <chris@m3.net>
 */

require_once MAX_PATH . '/lib/max/other/db_proc.php';
require_once MAX_PATH . '/lib/max/other/lib-db.inc.php';

/**
 * A function to log an ad requests.
 *
 * In order for the browser and operating system information to be
 * correctly logged, PHP must be set up correctly to utilise an up
 * to date browscap.ini file: see
 * {@link http://www.php.net/manual/en/ref.misc.php#ini.browscap}.
 *
 * @param integer $viewerId The viewer ID (was userid).
 * @param integer $adId The advertisement ID (was bannerid).
 * @param integer $creativeId The creative ID (doesn't exist yet, use null).
 * @param integer $zoneId The zone ID.
 */
function logAdRequest($viewerId, $adId, $creativeId, $zoneId)
{
    if (_viewersHostOkayToLog()) {
        $conf = $GLOBALS['_MAX']['CONF'];
        list($geotargeting, $zoneInfo, $userAgentInfo, $maxHttps) = _prepareLogInfo();
        $table = $conf['table']['prefix'] . $conf['table']['data_raw_ad_request'];
        _insertRawData($table, $viewerId, $adId, $creativeId,
                $zoneId, $geotargeting, $zoneInfo, $userAgentInfo, $maxHttps);
    }
}

/**
 * A function to log an ad impression.
 *
 * In order for the browser and operating system information to be
 * correctly logged, PHP must be set up correctly to utilise an up
 * to date browscap.ini file: see
 * {@link http://www.php.net/manual/en/ref.misc.php#ini.browscap}.
 *
 * @param integer $viewerId The viewer ID (was userid).
 * @param integer $adId The advertisement ID (was bannerid).
 * @param integer $creativeId The creative ID (doesn't exist yet, use null).
 * @param integer $zoneId The zone ID.
 */
function logAdImpression($viewerId, $adId, $creativeId, $zoneId)
{
    if (_viewersHostOkayToLog()) {
        $conf = $GLOBALS['_MAX']['CONF'];
        list($geotargeting, $zoneInfo, $userAgentInfo, $maxHttps) = _prepareLogInfo();
        $table = $conf['table']['prefix'] . $conf['table']['data_raw_ad_impression'];
        _insertRawData($table, $viewerId, $adId, $creativeId,
                $zoneId, $geotargeting, $zoneInfo, $userAgentInfo, $maxHttps);
    }
}

/**
 * A function to log an ad click.
 *
 * In order for the browser and operating system information to be
 * correctly logged, PHP must be set up correctly to utilise an up
 * to date browscap.ini file: see
 * {@link http://www.php.net/manual/en/ref.misc.php#ini.browscap}.
 *
 * @param integer $viewerId The viewer ID (was userid).
 * @param integer $adId The advertisement ID (was bannerid).
 * @param integer $creativeId The creative ID (doesn't exist yet, use null).
 * @param integer $zoneId The zone ID.
 */
function logAdClick($viewerId, $adId, $creativeId, $zoneId)
{
    if (_viewersHostOkayToLog()) {
        $conf = $GLOBALS['_MAX']['CONF'];
        list($geotargeting, $zoneInfo, $userAgentInfo, $maxHttps) = _prepareLogInfo();
        $table = $conf['table']['prefix'] . $conf['table']['data_raw_ad_click'];
        _insertRawData($table, $viewerId, $adId, $creativeId,
                $zoneId, $geotargeting, $zoneInfo, $userAgentInfo, $maxHttps);
    }
}

/**
 * A function to log a tracker impression.
 *
 * In order for the browser and operating system information to be
 * correctly logged, PHP must be set up correctly to utilise an up
 * to date browscap.ini file: see
 * {@link http://www.php.net/manual/en/ref.misc.php#ini.browscap}.
 *
 * Note that the $conf['rawDatabase'] variables will only be defined
 * in the event that Max is configured for multiple databases. Normally,
 * this will not be the case, so the server_ip field will be blank.
 *
 * @param integer $viewerId The viewer ID (was userid).
 * @param integer $trackerId The tracker ID.
 * @return mixed An array containing the server_raw_tracker_impression_id
 *               and the server_raw_ip values, if the tracker impression
 *               was inserted, false otherwise.
 */
function logTrackerImpression($viewerId, $trackerId)
{
    if (_viewersHostOkayToLog()) {
        $conf = $GLOBALS['_MAX']['CONF'];
        if (is_null($conf['rawDatabase']['host'])) {
            $conf['rawDatabase']['host'] = 'singleDB';
        }
        $serverRawIp = (isset($conf['rawDatabase']['serverRawIp'])) ? $conf['rawDatabase']['serverRawIp'] : $conf['rawDatabase']['host'];
            
        list($geotargeting, $zoneInfo, $userAgentInfo, $maxHttps) = _prepareLogInfo();
        if (phpAds_dbConnect(phpAds_rawDb)) {
            $dateFunc = !empty($conf['logging']['logInUTC']) ? 'gmdate' : 'date';
            $query = "
                INSERT INTO
                    {$conf['table']['prefix']}{$conf['table']['data_raw_tracker_impression']}
                    (
                        server_raw_ip,
                        viewer_id,
                        viewer_session_id,
                        date_time,
                        tracker_id,
                        channel,
                        language,
                        ip_address,
                        host_name,
                        country,
                        https,
                        domain,            
                        page,
                        query,
                        referer,
                        search_term,
                        user_agent,
                        os,
                        browser,
                        max_https
                    )
                VALUES
                    (
                        '$serverRawIp',
                        '$viewerId',
                        '',
                        '".$dateFunc('Y-m-d H:i:s')."',
                        '$trackerId',
                        '{$_GET['source']}',
                        '{$_SERVER['HTTP_ACCEPT_LANGUAGE']}',
                        '{$_SERVER['REMOTE_ADDR']}',
                        '{$_SERVER['REMOTE_HOST']}',
                        '{$geotargeting['country_code']}',
                        '{$zoneInfo['scheme']}',
                        '{$zoneInfo['host']}',
                        '{$zoneInfo['path']}',
                        '{$zoneInfo['query']}',
                        '{$_GET['referer']}',
                        '',
                        '{$_SERVER['HTTP_USER_AGENT']}',
                        '{$userAgentInfo['platform']}',
                        '{$userAgentInfo['browser']}',
                        '$maxHttps'
                    )";
            phpAds_dbQuery($query, phpAds_rawDb);
            return array('server_raw_tracker_impression_id' => phpAds_dbInsertID(phpAds_rawDb),
                         'server_raw_ip' => $serverRawIp);
        }
    }
    return false;
}

/**
 * A function to log a tracker impression variable values.
 *
 * Note that the $conf['rawDatabase'] variables will only be defined
 * in the event that Max is configured for multiple databases. Normally,
 * this will not be the case, so the server_ip field will be blank.
 *
 * @param integer $trackerId The tracker ID.
 * @param integer $serverRawTrackerImpressionId The unique tracker impression
 *                                              id on the raw database server.
 * @param string $serverRawIp The IP address of the raw database server, or null
 *                            if Max is not running in multiple database server
 *                            mode.
 */
function logVariableValues($trackerId, $serverRawTrackerImpressionId, $serverRawIp)
{
    $conf = $GLOBALS['_MAX']['CONF'];
    if (phpAds_dbConnect(phpAds_rawDb)) {
        // Get the variable information, including the Variable ID
        // (variableid) and the Variable Name (name)
        // $variables = MAX_getCacheVariablesByTrackerId($trackerId);
		$variables = MAX_getCachedEntities('variable', array('tracker_id' => $trackerId));
		$dateFunc = !empty($conf['logging']['logInUTC']) ? 'gmdate' : 'date';
        foreach ($variables as $variable) {
            if (isset($_GET[$variable['name']])) {
                $value = $_GET[$variable['name']];
            } else {
                $value = 'NULL';
            }
            $query = "
                INSERT INTO
                    {$conf['table']['prefix']}{$conf['table']['data_raw_tracker_variable_value']}
                    (
                        tracker_variable_id,
                        server_raw_tracker_impression_id,
                        server_raw_ip,
                        date_time,
                        value
                    )
                VALUES
                    (
                        '{$variable['variable_id']}',
                        '$serverRawTrackerImpressionId',
                        '$serverRawIp',
                        '".$dateFunc('Y-m-d H:i:s')."',
                        '$value'
                    )";
            phpAds_dbQuery($query, phpAds_rawDb);
        }
    }
}

/**
 * A function to log a tracker click.
 *
 * In order for the browser and operating system information to be
 * correctly logged, PHP must be set up correctly to utilise an up
 * to date browscap.ini file: see
 * {@link http://www.php.net/manual/en/ref.misc.php#ini.browscap}.
 *
 * @param integer $viewerId The viewer ID (was userid).
 * @param integer $trackerId The tracker ID.
 */
function logTrackerClick($viewerId, $trackerId)
{
    if (_viewersHostOkayToLog()) {
        $conf = $GLOBALS['_MAX']['CONF'];
        list($geotargeting, $zoneInfo, $userAgentInfo, $maxHttps) = _prepareLogInfo();
        if (phpAds_dbConnect(phpAds_rawDb)) {
            $dateFunc = !empty($conf['logging']['logInUTC']) ? 'gmdate' : 'date';
            $query = "
                INSERT INTO
                    {$conf['table']['prefix']}{$conf['table']['data_raw_tracker_click']}
                    (
                        viewer_id,
                        viewer_session_id,
                        date_time,
                        tracker_id,
                        channel,
                        language,
                        ip_address,
                        host_name,
                        country,
                        https,
                        domain,            
                        page,
                        query,
                        referer,
                        search_term,
                        user_agent,
                        os,
                        browser,
                        max_https
                    )
                VALUES
                    (
                        '$viewerId',
                        '',
                        '".$dateFunc('Y-m-d H:i:s')."',
                        '$trackerId',
                        '{$_GET['source']}',
                        '{$_SERVER['HTTP_ACCEPT_LANGUAGE']}',
                        '{$_SERVER['REMOTE_ADDR']}',
                        '{$_SERVER['REMOTE_HOST']}',
                        '{$geotargeting['country_code']}',
                        '{$zoneInfo['scheme']}',
                        '{$zoneInfo['host']}',
                        '{$zoneInfo['path']}',
                        '{$zoneInfo['query']}',
                        '{$_GET['referer']}',
                        '',
                        '{$_SERVER['HTTP_USER_AGENT']}',
                        '{$userAgentInfo['platform']}',
                        '{$userAgentInfo['browser']}',
                        '$maxHttps'
                    )";
            phpAds_dbQuery($query, phpAds_rawDb);
        }
    }
}

/**
 * A function to log benchmarking data to a file.
 * 
 * A file called "debug.log" is created in the cache directory to store
 * this information.
 * 
 * @param string $page The name of the script being benchmarked.
 * @param string $queryString The query string parameters passed to the page.
 * @param double $benchmark The elapsed time of the benchmark.
 * @param string $extra Any extra information to be logged.
 */
function logBenchmark($page, $queryString, $benchmark, $extra = '') {
    if (_viewersHostOkayToLog()) {
        $memoryUsage = function_exists('memory_get_usage') ? memory_get_usage() : '';
        $message = date("d/m/Y|H:i:s") . '|' . $memoryUsage . '|' . $page . '|' .
            $benchmark . '|' . $extra . '|' . $queryString . "\n";
        $logFile = fopen(MAX_PATH . '/var/debug.log', 'a');
        fwrite($logFile, $message);
        fclose($logFile);
    }
}

/**
 * A function to check if the information to be logged should be logged
 * or ignored, on the basis of the viewer's IP address or hostname.
 *
 * @return boolean True if the information should be logged, or false if the
 *                 host name is in the list of hosts to ignore log information
 *                 for.
 */
function _viewersHostOkayToLog()
{
    $conf = $GLOBALS['_MAX']['CONF'];
    if (count($conf['ignoreHosts']) > 0) {
        $hosts = '#('.implode('|',$conf['ignoreHosts']).')$#i';
        if ($hosts != '') {
            // Format the hosts to ignore in a PCRE format
			$hosts = str_replace('.', '\.', $hosts);
			$hosts = str_replace('*', '[^.]+', $hosts);
			// Check if the viewer's IP address is in the ignore list
			if (preg_match($hosts, $_SERVER['REMOTE_ADDR'])) {
				return false;
			}
			if ($conf['logging']['reverseLookup'] && is_null($_SERVER['REMOTE_HOST'])) {
			    // Set to reverse lookup IP addresses, and there is no
			    // hostname, so look up the hostname and set for future use
			    $hostname = gethostbyaddr($_SERVER['REMOTE_ADDR']);
			    $_SERVER['REMOTE_HOST'] = $hostname;
			} else {
			    // Logging just IP address, so use hostname if available,
			    // otherwise will just be null, and therefore skipped
                $hostname = $_SERVER['REMOTE_HOST'];
			}
			// Check if the viewer's hostname is in the ignore list
			if (preg_match($hosts, $hostname)) {
				return false;
			}
		}
	}
	return true;        
}

/**
 * A function to get various information that needs to be logged.
 *
 * @return array Returns an array with four elements:
 *                  0: An array of the viewer's geotargeting info.
 *                  1: An array of information about the URL the viewer used
 *                     to access the page containing the zone.
 *                  2: An array of information about the viewer's web browser
 *                     and operating system.
 *                  3: An integer to store if the call to Max was performed
 *                     using HTTPS (1) or not (0).
 */
function _prepareLogInfo()
{
    $conf = $GLOBALS['_MAX']['CONF'];
    // Set the viewer's hostname, if required
    if ($conf['logging']['reverseLookup'] && is_null($_SERVER['REMOTE_HOST'])) {
        // Set to reverse lookup IP addresses, and there is no
        // hostname, so look up the hostname and set for future use
        $hostname = gethostbyaddr($_SERVER['REMOTE_ADDR']);
        $_SERVER['REMOTE_HOST'] = $hostname;
    }
    // Get the Geotargeting information, if required
    $geotargeting = array();
    if ($conf['geotargeting']['type']) {
        $geotargeting = $GLOBALS['geoInfo'];
    }
    // Get the zone location information, if possible
    $zoneInfo = array();
    if (isset($_SERVER['HTTP_REFERER'])) {
        $zonePageURL = $_SERVER['HTTP_REFERER'];
    }
    if ((is_null($zonePageURL)) && isset($_GET['loc'])) {
        $zonePageURL = $_GET['loc'];
    }
    if (isset($zonePageURL)) {
        $zoneInfo = parse_url($zonePageURL);
        if (isset($zoneInfo)) {
            if ($zoneInfo['scheme'] == 'https') {
                $zoneInfo['scheme'] = 1;    // https is true
            } else {
                $zoneInfo['scheme'] = 0;    // https is false
            }
        }
    }
    // Get the operating system and browser type, if required
    $userAgentInfo = array();
    if ($conf['logging']['sniff'] && isset($_SERVER['HTTP_USER_AGENT'])) {
        include_once MAX_PATH . '/lib/phpSniff/phpSniff.class.php';
        $client = new phpSniff($_SERVER['HTTP_USER_AGENT']);
        $userAgentInfo['os'] = $client->property('os');
        $userAgentInfo['long_name'] = $client->property('long_name');
    }
    // Determine if the access to Max was made using HTTPS
    $maxHttps = 0;  // https is false
    if ($_SERVER['SERVER_PORT'] == $conf['max']['sslPort']) {
        $maxHttps = 1;   // https is true
    }
    return array($geotargeting, $zoneInfo, $userAgentInfo, $maxHttps);
}

/**
 * A function to insert ad requests, ad impressions, ad clicks
 * and tracker clicks into the raw tables. Does NOT work with
 * tracker impressions.
 *
 * @param string $table A reference to the raw table name to insert into.
 * @param integer $viewerId A reference to the viewer ID.
 * @param integer $adId A reference to the advertisement ID.
 * @param integer $creativeId A reference to the creative ID.
 * @param integer $zoneId A reference to the zone ID.
 * @param array $geotargeting A reference to an array holding the viewer's geotargeting info.
 * @param array $zoneInfo A reference to an array to store information about the URL 
 *                        the viewer used to access the page containing the zone.
 * @param array $userAgentInfo A reference to an array to store information about the
 *                             viewer's web browser and operating system.
 * @param integer $maxHttps A reference to an integer to store if the call to Max was
 *                          performed using HTTPS or not.
 */
function _insertRawData(&$table, &$viewerId, &$adId, &$creativeId, &$zoneId,
                        &$geotargeting, &$zoneInfo, &$userAgentInfo, &$maxHttps)
{
    if (phpAds_dbConnect(phpAds_rawDb)) {
        $dateFunc = !empty($conf['logging']['logInUTC']) ? 'gmdate' : 'date';
        $query = "
            INSERT INTO
                $table
                (
                    viewer_id,
                    viewer_session_id,
                    date_time,
                    ad_id,
                    creative_id,
                    zone_id,
                    channel,
                    language,
                    ip_address,
                    host_name,
                    country,
                    https,
                    domain,            
                    page,
                    query,
                    referer,
                    search_term,
                    user_agent,
                    os,
                    browser,
                    max_https,
                    geo_region,
                    geo_city,
                    geo_postal_code,
                    geo_latitude,
                    geo_longitude,
                    geo_dma_code,
                    geo_area_code,
                    geo_organisation,
                    geo_netspeed,
                    geo_continent
                )
            VALUES
                (
                    '$viewerId',
                    '',
                    '".$dateFunc('Y-m-d H:i:s')."',
                    '$adId',
                    '$creativeId',
                    '$zoneId',
                    '{$_GET['source']}',
                    '{$_SERVER['HTTP_ACCEPT_LANGUAGE']}',
                    '{$_SERVER['REMOTE_ADDR']}',
                    '{$_SERVER['REMOTE_HOST']}',
                    '{$geotargeting['country_code']}',
                    '{$zoneInfo['scheme']}',
                    '{$zoneInfo['host']}',
                    '{$zoneInfo['path']}',
                    '{$zoneInfo['query']}',
                    '{$_GET['referer']}',
                    '',
                    '{$_SERVER['HTTP_USER_AGENT']}',
                    '{$userAgentInfo['os']}',
                    '{$userAgentInfo['long_name']}',
                    '$maxHttps',
                    '{$geotargeting['region']}',
                    '{$geotargeting['city']}',
                    '{$geotargeting['postal_code']}',
                    '{$geotargeting['latitude']}',
                    '{$geotargeting['longitude']}',
                    '{$geotargeting['dma_code']}',
                    '{$geotargeting['area_code']}',
                    '{$geotargeting['organisation']}',
                    '{$geotargeting['netspeed']}',
                    '{$geotargeting['continent']}'
                )";
        phpAds_dbQuery($query, phpAds_rawDb);
    }
}

?>